# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Sets up and handles the text editor conext menus"""


import os.path
from cdmpyparser import getBriefModuleInfoFromMemory
from ui.qt import QMenu, QActionGroup, QApplication, QMessageBox
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.encoding import (SUPPORTED_CODECS,
                            getNormalizedEncoding,
                            detectEncodingOnClearExplicit,
                            detectNewFileWriteEncoding)
from utils.diskvaluesrelay import getFileEncoding, setFileEncoding
from autocomplete.bufferutils import getContext
from analysis.disasm import (OPT_NO_OPTIMIZATION, OPT_OPTIMIZE_ASSERT,
                             OPT_OPTIMIZE_DOCSTRINGS)


class EditorContextMenuMixin:

    """Encapsulates the context menu handling"""

    def __init__(self):
        self.encodingReloadMenu = QMenu("Set &encoding and reload")
        self.encodingReloadActGrp = QActionGroup(self)
        self.encodingWriteMenu = QMenu("Set encodin&g")
        self.encodingWriteActGrp = QActionGroup(self)

        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager

        self._menu = QMenu(self)
        self.__menuUndo = self._menu.addAction(
            getIcon('undo.png'), '&Undo', self.onUndo, "Ctrl+Z")
        self.__menuRedo = self._menu.addAction(
            getIcon('redo.png'), '&Redo', self.onRedo, "Ctrl+Y")
        self._menu.addSeparator()

        self.__menuCut = self._menu.addAction(
            getIcon('cutmenu.png'), 'Cu&t', self.onShiftDel, "Ctrl+X")
        self.__menuCopy = self._menu.addAction(
            getIcon('copymenu.png'), '&Copy', self.onCtrlC, "Ctrl+C")
        self.__menuPaste = self._menu.addAction(
            getIcon('pastemenu.png'), '&Paste', self.paste, "Ctrl+V")
        self.__menuSelectAll = self._menu.addAction(
            getIcon('selectallmenu.png'), 'Select &all',
            self.selectAll, "Ctrl+A")
        self._menu.addSeparator()

        self.__initReloadEncodingMenu()
        self.encodingReloadMenu.setIcon(getIcon('textencoding.png'))
        self._menu.addMenu(self.encodingReloadMenu)
        self.__initWriteEncodingMenu()
        self.encodingWriteMenu.setIcon(getIcon('textencoding.png'))
        menu = self._menu.addMenu(self.encodingWriteMenu)
        self.__menuClearEncoding = self._menu.addAction(
            getIcon('clearmenu.png'), 'Clear explicit encoding',
            self.__onClearEncoding)
        self._menu.addSeparator()

        menu = self._menu.addMenu(self.__initToolsMenu())
        menu.setIcon(getIcon('toolsmenu.png'))
        self._menu.addSeparator()

        menu = self._menu.addMenu(self.__initDiagramsMenu())
        menu.setIcon(getIcon('diagramsmenu.png'))
        self._menu.addSeparator()

        self.__menuOpenAsFile = self._menu.addAction(
            getIcon('filemenu.png'), 'O&pen as file', self.openAsFile)
        self.__menuDownloadAndShow = self._menu.addAction(
            getIcon('filemenu.png'), 'Do&wnload and show',
            self.downloadAndShow)
        self.__menuOpenInBrowser = self._menu.addAction(
            getIcon('homepagemenu.png'), 'Open in browser', self.openInBrowser)
        self._menu.addSeparator()

        self._menuHighlightInPrj = self._menu.addAction(
            getIcon("highlightmenu.png"), "&Highlight in project browser",
            editorsManager.onHighlightInPrj)
        self._menuHighlightInFS = self._menu.addAction(
            getIcon("highlightmenu.png"), "H&ighlight in file system browser",
            editorsManager.onHighlightInFS)
        self._menuHighlightInOutline = self._menu.addAction(
            getIcon("highlightmenu.png"), "Highlight in &outline browser",
            self.highlightInOutline, 'Ctrl+B')

        # Plugins support
        self.__pluginMenuSeparator = self._menu.addSeparator()
        registeredMenus = editorsManager.getPluginMenus()
        if registeredMenus:
            for path in registeredMenus:
                self._menu.addMenu(registeredMenus[path])
        else:
            self.__pluginMenuSeparator.setVisible(False)

        editorsManager.sigPluginContextMenuAdded.connect(
            self._onPluginMenuAdded)
        editorsManager.sigPluginContextMenuRemoved.connect(
            self._onPluginMenuRemoved)

    def __initReloadEncodingMenu(self):
        """Creates the encoding menu for reloading the existing file"""
        for encoding in sorted(SUPPORTED_CODECS):
            act = self.encodingReloadMenu.addAction(encoding)
            act.setCheckable(True)
            act.setData(encoding)
            self.encodingReloadActGrp.addAction(act)
        self.encodingReloadMenu.triggered.connect(self.__onReloadWithEncoding)

    def __initWriteEncodingMenu(self):
        """Creates the encoding menu for further read/write operations"""
        for encoding in sorted(SUPPORTED_CODECS):
            act = self.encodingWriteMenu.addAction(encoding)
            act.setCheckable(True)
            act.setData(encoding)
            self.encodingWriteActGrp.addAction(act)
        self.encodingWriteMenu.triggered.connect(self.__onReadWriteEncoding)

    def __initToolsMenu(self):
        """Creates the tools menu"""
        self.toolsMenu = QMenu('Python too&ls')
        self.runAct = self.toolsMenu.addAction(
            getIcon('run.png'), 'Run script', self._parent.onRunScript)
        self.runParamAct = self.toolsMenu.addAction(
            getIcon('paramsmenu.png'), 'Set parameters and run',
            self._parent.onRunScriptDlg)
        self.toolsMenu.addSeparator()
        self.profileAct = self.toolsMenu.addAction(
            getIcon('profile.png'), 'Profile script',
            self._parent.onProfileScript)
        self.profileParamAct = self.toolsMenu.addAction(
            getIcon('paramsmenu.png'), 'Set parameters and profile',
            self._parent.onProfileScriptDlg)
        self.toolsMenu.addSeparator()
        self.disasmMenu = QMenu('Disassembly')
        self.disasmMenu.setIcon(getIcon('disassembly.png'))
        self.disasmAct0 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (no optimization)',
            self._onDisasm0)
        self.disasmAct1 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 1)',
            self._onDisasm1)
        self.disasmAct2 = self.disasmMenu.addAction(
            getIcon(''), 'Disassembly (optimization level 2)',
            self._onDisasm2)
        self.toolsMenu.addMenu(self.disasmMenu)
        return self.toolsMenu

    def __initDiagramsMenu(self):
        """Creates the diagrams menu"""
        self.diagramsMenu = QMenu("&Diagrams")
        self.importsDgmAct = self.diagramsMenu.addAction(
            getIcon('importsdiagram.png'), 'Imports diagram',
            self._parent.onImportDgm)
        self.importsDgmParamAct = self.diagramsMenu.addAction(
            getIcon('paramsmenu.png'), 'Fine tuned imports diagram',
            self._parent.onImportDgmTuned)
        return self.diagramsMenu

    def contextMenuEvent(self, event):
        """Called just before showing a context menu"""
        # Accepting needs to suppress the native menu
        event.accept()

        isPython = self.isPythonBuffer()
        readOnly = self.isReadOnly()
        self.__menuUndo.setEnabled(self.document().isUndoAvailable())
        self.__menuRedo.setEnabled(self.document().isRedoAvailable())
        self.__menuCut.setEnabled(not readOnly)
        self.__menuPaste.setEnabled(QApplication.clipboard().text() != ""
                                    and not readOnly)

        fileName = self._parent.getFileName()
        absFileName = os.path.isabs(fileName)
        self.__menuOpenAsFile.setEnabled(self.openAsFileAvailable())
        self.__menuDownloadAndShow.setEnabled(self.downloadAndShowAvailable())
        self.__menuOpenInBrowser.setEnabled(self.downloadAndShowAvailable())
        self._menuHighlightInPrj.setEnabled(
            absFileName and GlobalData().project.isProjectFile(fileName))
        self._menuHighlightInFS.setEnabled(absFileName)
        self._menuHighlightInOutline.setEnabled(isPython)
        self._menuHighlightInOutline.setEnabled(isPython)

        self.toolsMenu.setEnabled(isPython)
        if isPython:
            runEnabled = self._parent.runScriptButton.isEnabled()
            self.runAct.setEnabled(runEnabled)
            self.runParamAct.setEnabled(runEnabled)
            self.profileAct.setEnabled(runEnabled)
            self.profileParamAct.setEnabled(runEnabled)

        if absFileName:
            self.__menuClearEncoding.setEnabled(
                getFileEncoding(fileName) is not None)
        else:
            self.__menuClearEncoding.setEnabled(
                self.explicitUserEncoding is not None)

        # Check the proper encoding in the menu
        encoding = 'undefined'
        if absFileName:
            enc = getFileEncoding(fileName)
            if enc:
                encoding = enc
        else:
            if self.explicitUserEncoding:
                encoding = self.explicitUserEncoding
        encoding = getNormalizedEncoding(encoding, False)
        if absFileName:
            for act in self.encodingReloadActGrp.actions():
                act.setChecked(encoding == getNormalizedEncoding(act.data()))
        else:
            self.encodingReloadMenu.setEnabled(False)
        for act in self.encodingWriteActGrp.actions():
            act.setChecked(encoding == getNormalizedEncoding(act.data()))

        # Show the menu
        self._menu.popup(event.globalPos())

    def __isSameEncodingAsCurrent(self, enc):
        """True if the same encoding has already been set"""
        fileName = self._parent.getFileName()
        if not os.path.isabs(fileName):
            # New unsaved yet file
            if not self.explicitUserEncoding:
                return False
            return getNormalizedEncoding(enc) == getNormalizedEncoding(
                self.explicitUserEncoding)

        # Existed before or just saved new file
        currentEnc = getFileEncoding(fileName)
        if not currentEnc:
            return False
        return getNormalizedEncoding(currentEnc) == getNormalizedEncoding(enc)

    def __onReloadWithEncoding(self, act):
        """Triggered when encoding is selected"""
        # The method is called only for the existing disk files
        encoding = act.data()
        if self.__isSameEncodingAsCurrent(encoding):
            return

        if self.document().isModified():
            res = QMessageBox.warning(
                self, 'Continue loosing changes',
                '<p>The buffer has unsaved changes. Are you sure to continue '
                'reloading the content using ' + encoding + ' encoding and '
                'loosing the changes?</p>',
                QMessageBox.StandardButtons(QMessageBox.Cancel |
                                            QMessageBox.Yes),
                QMessageBox.Cancel)
            if res == QMessageBox.Cancel:
                return

        # Do the reload
        fileName = self._parent.getFileName()
        setFileEncoding(fileName, encoding)
        self.__updateFilePosition()
        self.readFile(fileName)
        self.__restoreFilePosition()
        self.__updateMainWindowStatusBar()

    def __onReadWriteEncoding(self, act):
        """Sets explicit encoding for further read/write ops"""
        encoding = act.data()
        if self.__isSameEncodingAsCurrent(encoding):
            return

        # fileName = self._parent.getFileName()
        # absFileName = os.path.isabs(fileName)
        self.document().setModified(True)
        self.explicitUserEncoding = encoding
        self.__updateMainWindowStatusBar()

    def __onClearEncoding(self):
        """Clears the explicitly set encoding"""
        self.explicitUserEncoding = None
        fileName = self._parent.getFileName()
        absFileName = os.path.isabs(fileName)
        if absFileName:
            setFileEncoding(fileName, None)
            self.encoding = detectEncodingOnClearExplicit(fileName, self.text)
        self.__updateMainWindowStatusBar()

    def onUndo(self):
        """Undo implementation"""
        if self.document().isUndoAvailable():
            self.undo()
            self._parent.modificationChanged()

    def onRedo(self):
        """Redo implementation"""
        if self.document().isRedoAvailable():
            self.redo()
            self._parent.modificationChanged()

    def _onPluginMenuAdded(self, menu, count):
        """Triggered when a new menu was added"""
        del count   # unused argument
        self._menu.addMenu(menu)
        self.__pluginMenuSeparator.setVisible(True)

    def __onDisasm(self, optimization):
        """Common implementation"""
        if self.isPythonBuffer():
            if os.path.isabs(self._parent.getFileName()):
                if not self._parent.isModified():
                    GlobalData().mainWindow.showFileDisassembly(
                        self._parent.getFileName(), optimization)
                    return
            fileName = self._parent.getFileName()
            if not fileName:
                fileName = self._parent.getShortName()
            encoding = self.encoding
            if not encoding:
                encoding = detectNewFileWriteEncoding(self, fileName)
            GlobalData().mainWindow.showBufferDisassembly(
                self.text, encoding, fileName, optimization)

    def _onDisasm0(self):
        """Triggered to disassemble the buffer without optimization"""
        self.__onDisasm(OPT_NO_OPTIMIZATION)

    def _onDisasm1(self):
        """Triggered to disassemble the buffer with optimization level 1"""
        self.__onDisasm(OPT_OPTIMIZE_ASSERT)

    def _onDisasm2(self):
        """Triggered to disassemble the buffer with optimization level 2"""
        self.__onDisasm(OPT_OPTIMIZE_DOCSTRINGS)

    def _onPluginMenuRemoved(self, menu, count):
        """Triggered when a menu was deleted"""
        self._menu.removeAction(menu.menuAction())
        self.__pluginMenuSeparator.setVisible(count != 0)

    def highlightInOutline(self):
        """Triggered when highlight in outline browser is requested"""
        if self.isPythonBuffer():
            info = getBriefModuleInfoFromMemory(self.text)
            context = getContext(self, info, True, False)
            line, _ = self.cursorPosition
            GlobalData().mainWindow.highlightInOutline(context, int(line) + 1)
            self.setFocus()

    def terminateMenus(self):
        """Called when the tab is closed"""
        self.encodingReloadMenu.triggered.disconnect(
            self.__onReloadWithEncoding)
        self.encodingReloadMenu.deleteLater()
        self.encodingReloadActGrp.deleteLater()
        self.encodingWriteMenu.triggered.disconnect(self.__onReadWriteEncoding)
        self.encodingWriteMenu.deleteLater()
        self.encodingWriteActGrp.deleteLater()
        self.toolsMenu.deleteLater()
        self.disasmMenu.deleteLater()
        self.diagramsMenu.deleteLater()
        self._menu.deleteLater()

        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.sigPluginContextMenuAdded.disconnect(
            self._onPluginMenuAdded)
        editorsManager.sigPluginContextMenuRemoved.disconnect(
            self._onPluginMenuRemoved)

    @staticmethod
    def __updateMainWindowStatusBar():
        """Updates the main window status bar"""
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.updateStatusBar()

    @staticmethod
    def __updateFilePosition():
        """Updates the position in a file"""
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.updateFilePosition(None)

    @staticmethod
    def __restoreFilePosition():
        """Restores the position in a file"""
        mainWindow = GlobalData().mainWindow
        editorsManager = mainWindow.editorsManagerWidget.editorsManager
        editorsManager.restoreFilePosition(None)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Find and replace widgets implementation"""

import re
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from utils.settings import Settings
from utils.project import CodimensionProject
from utils.diskvaluesrelay import getFindHistory, setFindHistory
from utils.colorfont import setLineEditBackground
from .qt import (QToolButton, QLabel, QSizePolicy, QComboBox,
                 QGridLayout, QWidget, QCheckBox, QKeySequence, Qt, QSize,
                 QEvent, pyqtSignal)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class ComboBoxNoUndo(QComboBox):

    """Combo box which allows application wide Ctrl+Z etc."""

    sigNext = pyqtSignal()
    sigPrevious = pyqtSignal()
    sigEnter = pyqtSignal(int)

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)
        self.setSizeAdjustPolicy(QComboBox.AdjustToMinimumContentsLength)
        self.setEditable(True)

    def event(self, event):
        """Skips the Undo/Redo shortcuts. They're processed by the editor"""
        if event.type() == QEvent.ShortcutOverride:
            if event.matches(QKeySequence.Undo):
                return False
            if event.matches(QKeySequence.Redo):
                return False
        if event.type() == QEvent.KeyPress:
            key = event.key()
            modifiers = event.modifiers()
            if key in [Qt.Key_Enter, Qt.Key_Return]:
                self.sigEnter.emit(int(modifiers))
                return False
            if modifiers == Qt.ControlModifier:
                if key == Qt.Key_Comma:
                    self.sigPrevious.emit()
                    return False
                if key == Qt.Key_Period:
                    self.sigNext.emit()
                    return False
        return QComboBox.event(self, event)


class FindReplaceWidget(QWidget):

    """Find and replace widgets"""

    MODE_FIND = 0
    MODE_REPLACE = 1

    # Background colors
    BG_IDLE = 0
    BG_NOMATCH = 1
    BG_MATCH = 2
    BG_BROKEN = 3

    def __init__(self, editorsManager, parent=None):
        QWidget.__init__(self, parent)

        self.__maxEntries = Settings()['maxSearchEntries']

        self.editorsManager = editorsManager
        self.__editor = None
        self.__subscribedToCursor = False

        self.__createLayout()
        self.__changesConnected = False
        self.__connectOnChanges()

        self.__history = getFindHistory()
        self.__populateHistory()

        # Incremental search support
        self.__startPoint = None    # {'absPos': int, 'firstVisible': int}

        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

        lineEdit = self.findtextCombo.lineEdit()
        self.__stateColors = {
            self.BG_IDLE:
                lineEdit.palette().color(lineEdit.backgroundRole()),
            self.BG_NOMATCH:
                GlobalData().skin['findNoMatchPaper'],
            self.BG_MATCH:
                GlobalData().skin['findMatchPaper'],
            self.BG_BROKEN:
                GlobalData().skin['findInvalidPaper']}

    def __setBackgroundColor(self, state):
        """Sets the search combo background color to reflect the state"""
        setLineEditBackground(self.findtextCombo.lineEdit(),
                              self.__stateColors[state],
                              self.__stateColors[self.BG_IDLE])

    def __createLayout(self):
        """Creates the layout of the widget"""
        self.closeButton = QToolButton(self)
        self.closeButton.setToolTip("Close the dialog (ESC)")
        self.closeButton.setIcon(getIcon("close.png"))
        self.closeButton.clicked.connect(self.hide)

        self.findLabel = QLabel("Find:", self)

        self.findtextCombo = ComboBoxNoUndo(self)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.findtextCombo.sizePolicy().hasHeightForWidth())
        self.findtextCombo.setSizePolicy(sizePolicy)
        self.findtextCombo.setInsertPolicy(QComboBox.NoInsert)
        self.findtextCombo.setCompleter(None)
        self.findtextCombo.setDuplicatesEnabled(False)
        self.findtextCombo.setEnabled(False)
        self.findtextCombo.sigNext.connect(self.onNext)
        self.findtextCombo.sigPrevious.connect(self.onPrev)
        self.findtextCombo.sigEnter.connect(self.__onFindEnter)

        self.findPrevButton = QToolButton(self)
        self.findPrevButton.setToolTip("Previous match (Ctrl+,)")
        self.findPrevButton.setIcon(getIcon("1leftarrow.png"))
        self.findPrevButton.setIconSize(QSize(24, 16))
        self.findPrevButton.setFocusPolicy(Qt.NoFocus)
        self.findPrevButton.setEnabled(False)
        self.findPrevButton.clicked.connect(self.onPrev)

        self.findNextButton = QToolButton(self)
        self.findNextButton.setToolTip("Next match (Ctrl+.)")
        self.findNextButton.setIcon(getIcon("1rightarrow.png"))
        self.findNextButton.setIconSize(QSize(24, 16))
        self.findNextButton.setFocusPolicy(Qt.NoFocus)
        self.findNextButton.setEnabled(False)
        self.findNextButton.clicked.connect(self.onNext)

        self.caseCheckBox = QCheckBox(self)
        self.caseCheckBox.setText("Match case")
        self.caseCheckBox.setFocusPolicy(Qt.NoFocus)
        self.caseCheckBox.setEnabled(False)

        self.wordCheckBox = QCheckBox(self)
        self.wordCheckBox.setText("Whole word")
        self.wordCheckBox.setFocusPolicy(Qt.NoFocus)
        self.wordCheckBox.setEnabled(False)

        self.regexpCheckBox = QCheckBox(self)
        self.regexpCheckBox.setText("Regexp")
        self.regexpCheckBox.setFocusPolicy(Qt.NoFocus)
        self.regexpCheckBox.setEnabled(False)


        # Replace UI elements
        self.replaceLabel = QLabel("Replace:", self)

        self.replaceCombo = ComboBoxNoUndo(self)

        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.replaceCombo.sizePolicy().hasHeightForWidth())
        self.replaceCombo.setSizePolicy(sizePolicy)
        self.replaceCombo.setInsertPolicy(QComboBox.NoInsert)
        self.replaceCombo.setCompleter(None)
        self.replaceCombo.setDuplicatesEnabled(False)
        self.replaceCombo.setEnabled(False)
        self.replaceCombo.sigNext.connect(self.onNext)
        self.replaceCombo.sigPrevious.connect(self.onPrev)
        self.replaceCombo.sigEnter.connect(self.__onReplaceEnter)

        self.replaceButton = QToolButton(self)
        self.replaceButton.setToolTip("Replace current match")
        self.replaceButton.setIcon(getIcon("replace.png"))
        self.replaceButton.setEnabled(False)
        self.replaceButton.setFocusPolicy(Qt.NoFocus)
        self.replaceButton.clicked.connect(self.__onReplace)
        self.replaceButton.setIconSize(QSize(24, 16))

        self.replaceAllButton = QToolButton(self)
        self.replaceAllButton.setToolTip("Replace all matches")
        self.replaceAllButton.setIcon(getIcon("replace-all.png"))
        self.replaceAllButton.setIconSize(QSize(24, 16))
        self.replaceAllButton.setEnabled(False)
        self.replaceAllButton.setFocusPolicy(Qt.NoFocus)
        self.replaceAllButton.clicked.connect(self.__onReplaceAll)

        self.replaceAndMoveButton = QToolButton(self)
        self.replaceAndMoveButton.setToolTip(
            "Replace current match and move to the next one")
        self.replaceAndMoveButton.setIcon(getIcon("replace-move.png"))
        self.replaceAndMoveButton.setIconSize(QSize(24, 16))
        self.replaceAndMoveButton.setEnabled(False)
        self.replaceAndMoveButton.setFocusPolicy(Qt.NoFocus)
        self.replaceAndMoveButton.clicked.connect(self.__onReplaceAndMove)

        # Layout
        self.gridLayout = QGridLayout(self)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)

        self.gridLayout.addWidget(self.closeButton, 0, 0, 1, 1)
        self.gridLayout.addWidget(self.findLabel, 0, 1, 1, 1)
        self.gridLayout.addWidget(self.findtextCombo, 0, 2, 1, 1)
        self.gridLayout.addWidget(self.findPrevButton, 0, 3, 1, 1)
        self.gridLayout.addWidget(self.findNextButton, 0, 4, 1, 1)
        self.gridLayout.addWidget(self.caseCheckBox, 0, 5, 1, 1)
        self.gridLayout.addWidget(self.wordCheckBox, 0, 6, 1, 1)
        self.gridLayout.addWidget(self.regexpCheckBox, 0, 7, 1, 1)

        self.gridLayout.addWidget(self.replaceLabel, 1, 1, 1, 1)
        self.gridLayout.addWidget(self.replaceCombo, 1, 2, 1, 1)
        self.gridLayout.addWidget(self.replaceButton, 1, 3, 1, 1)
        self.gridLayout.addWidget(self.replaceAndMoveButton, 1, 4, 1, 1)
        self.gridLayout.addWidget(self.replaceAllButton, 1, 5, 1, 1)

        self.setTabOrder(self.findtextCombo, self.replaceCombo)
        self.setTabOrder(self.replaceCombo, self.caseCheckBox)
        self.setTabOrder(self.caseCheckBox, self.wordCheckBox)
        self.setTabOrder(self.wordCheckBox, self.regexpCheckBox)
        self.setTabOrder(self.regexpCheckBox, self.findNextButton)
        self.setTabOrder(self.findNextButton, self.findPrevButton)
        self.setTabOrder(self.findPrevButton, self.replaceAndMoveButton)
        self.setTabOrder(self.replaceButton, self.replaceAllButton)
        self.setTabOrder(self.replaceAndMoveButton, self.replaceAllButton)
        self.setTabOrder(self.replaceAllButton, self.closeButton)

    def keyPressEvent(self, event):
        """Handles the ESC key for the search bar"""
        if event.key() == Qt.Key_Escape:
            self.__startPoint = None
            event.accept()
            self.hide()
            activeWindow = self.editorsManager.currentWidget()
            if activeWindow:
                activeWindow.setFocus()

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__history = getFindHistory()
            self.__populateHistory()

    def __serialize(self):
        """Sirializes a current search/replace parameters"""
        termText = self.findtextCombo.currentText()
        if self.replaceCombo.isVisible():
            replaceText = self.replaceCombo.currentText()
        else:
            replaceText = ''

        return {'term': termText,
                'replace': replaceText,
                'cbCase': self.caseCheckBox.isChecked(),
                'cbWord': self.wordCheckBox.isChecked(),
                'cbRegexp': self.regexpCheckBox.isChecked()}

    def __deserialize(self, item):
        """Deserializes the history item"""
        self.__disconnectOnChanges()
        self.findtextCombo.setEditText(item['term'])
        self.replaceCombo.setEditText(item['replace'])
        self.caseCheckBox.setChecked(item['cbCase'])
        self.wordCheckBox.setChecked(item['cbWord'])
        self.regexpCheckBox.setChecked(item['cbRegexp'])
        self.__connectOnChanges()

    def __connectOnChanges(self):
        """Connects all the UI controls to their on change handlers"""
        if not self.__changesConnected:
            self.findtextCombo.editTextChanged.connect(
                self.__onCriteriaChanged)
            self.findtextCombo.currentIndexChanged[int].connect(
                self.__onfindIndexChanged)
            self.caseCheckBox.stateChanged.connect(self.__onCriteriaChanged)
            self.wordCheckBox.stateChanged.connect(self.__onCriteriaChanged)
            self.regexpCheckBox.stateChanged.connect(self.__onCriteriaChanged)
            self.__changesConnected = True

    def __disconnectOnChanges(self):
        """Disconnects all the UI controls from their on change handlers"""
        if self.__changesConnected:
            self.findtextCombo.editTextChanged.disconnect(
                self.__onCriteriaChanged)
            self.findtextCombo.currentIndexChanged[int].disconnect(
                self.__onfindIndexChanged)
            self.caseCheckBox.stateChanged.disconnect(self.__onCriteriaChanged)
            self.wordCheckBox.stateChanged.disconnect(self.__onCriteriaChanged)
            self.regexpCheckBox.stateChanged.disconnect(
                self.__onCriteriaChanged)
            self.__changesConnected = False

    def __populateHistory(self):
        """Populates the history"""
        self.__disconnectOnChanges()
        index = 0
        replaceItems = []
        for props in self.__history:
            self.findtextCombo.addItem(props['term'], index)
            replaceItem = props['replace']
            if replaceItem and replaceItem not in replaceItems:
                self.replaceCombo.addItem(props['replace'])
                replaceItems.append(props['replace'])
            index += 1
        self.replaceCombo.addItem('')
        self.__connectOnChanges()

    def __historyIndexByFindText(self, text):
        """Provides the history index by the find text"""
        if text:
            for index in range(self.findtextCombo.count()):
                if self.findtextCombo.itemText(index) == text:
                    return index, self.findtextCombo.itemData(index)
        return None, None

    def __replaceIndex(self, text):
        """Provides the replace combo index by the text"""
        if text is not None:
            for index in range(self.replaceCombo.count()):
                if self.replaceCombo.itemText(index) == text:
                    return index
        return None

    def __updateHistory(self):
        """Updates history if needed"""
        # Add entries to the combo box if required
        self.__disconnectOnChanges()
        currentText = self.findtextCombo.currentText()
        historyItem = self.__serialize()
        _, historyIndex = self.__historyIndexByFindText(currentText)
        if historyIndex is not None:
            self.__history[historyIndex] = historyItem
        else:
            self.__history.insert(0, historyItem)
            if len(self.__history) > self.__maxEntries:
                self.__history = self.__history[:self.__maxEntries]

        self.findtextCombo.clear()
        self.replaceCombo.clear()
        self.__populateHistory()

        self.__disconnectOnChanges()
        comboIndex, historyIndex = self.__historyIndexByFindText(currentText)
        self.findtextCombo.setCurrentIndex(comboIndex)

        replaceText = self.__history[historyIndex]['replace']
        replaceIndex = self.__replaceIndex(replaceText)
        if replaceIndex:
            self.replaceCombo.setCurrentIndex(replaceIndex)
        else:
            self.replaceCombo.setEditText(
                self.__history[historyIndex]['replace'])
        self.__connectOnChanges()

        # Save the combo values for further usage
        setFindHistory(self.__history)

    def __disableAll(self):
        """Disables all the controls"""
        self.findtextCombo.setEnabled(False)
        self.caseCheckBox.setEnabled(False)
        self.wordCheckBox.setEnabled(False)
        self.regexpCheckBox.setEnabled(False)
        self.findPrevButton.setEnabled(False)
        self.findNextButton.setEnabled(False)
        self.replaceButton.setEnabled(False)
        self.replaceAndMoveButton.setEnabled(False)
        self.replaceAllButton.setEnabled(False)
        self.replaceCombo.setEnabled(False)
        self.__setBackgroundColor(self.BG_IDLE)

    def updateStatus(self):
        """Triggered when the current tab is changed"""
        self.__unsubscribeFromEditorSignals()
        currentWidget = self.editorsManager.currentWidget()
        validWidgets = [MainWindowTabWidgetBase.PlainTextEditor,
                        MainWindowTabWidgetBase.VCSAnnotateViewer]
        if currentWidget.getType() not in validWidgets:
            self.__editor = None
            self.__disableAll()
            return

        self.__editor = currentWidget.getEditor()
        self.__editor.onTabChanged()
        self.__startPoint = None

        self.findtextCombo.setEnabled(True)
        self.caseCheckBox.setEnabled(True)
        self.wordCheckBox.setEnabled(True)
        self.regexpCheckBox.setEnabled(True)

        criteriaValid = self.__isCriteriaValid()
        if criteriaValid:
            _, totalMatches = self.__editor.getMatchesInfo()
            if totalMatches is None or totalMatches == 0:
                self.__setBackgroundColor(self.BG_NOMATCH)
            else:
                self.__setBackgroundColor(self.BG_MATCH)
        else:
            self.__setBackgroundColor(self.BG_BROKEN)

        self.findPrevButton.setEnabled(criteriaValid)
        self.findNextButton.setEnabled(criteriaValid)

        self.replaceCombo.setEnabled(True)
        self.replaceButton.setEnabled(False)
        self.replaceAndMoveButton.setEnabled(False)
        self.replaceAllButton.setEnabled(False)
        self.__subscribeToEditorSignals()

    def __isCriteriaValid(self):
        """True if the search criteria is valid"""
        if self.findtextCombo.currentText() != "":
            valid, _ = self.__isSearchRegexpValid()
            if valid:
                return True
        return False

    def show(self, mode, text=''):
        """Overridden show method"""
        self.__disconnectOnChanges()
        self.findtextCombo.clear()
        self.replaceCombo.clear()
        self.regexpCheckBox.setChecked(False)
        self.caseCheckBox.setChecked(False)
        self.wordCheckBox.setChecked(False)

        self.__populateHistory()

        self.__disconnectOnChanges()
        self.findtextCombo.setEditText(text)
        if mode == self.MODE_FIND:
            self.findtextCombo.lineEdit().selectAll()
        self.replaceCombo.setEditText('')
        self.replaceCombo.setCurrentIndex(self.__replaceIndex(''))
        self.__connectOnChanges()

        replaceVisible = mode == self.MODE_REPLACE
        self.replaceLabel.setVisible(replaceVisible)
        self.replaceCombo.setVisible(replaceVisible)
        self.replaceButton.setVisible(replaceVisible)
        self.replaceAndMoveButton.setVisible(replaceVisible)
        self.replaceAllButton.setVisible(replaceVisible)

        QWidget.show(self)
        self.activateWindow()

        if self.__editor is not None:
            # Even if the cursor is in a middle of a word, the Ctrl+F does not
            # move to the next match. It is done in the editorsmanager handler.
            # When a search is initiated the cursor is moved to the begiining
            # of a word if so. Thus, here we always start to search forward
            # without jumping to the next match.
            self.__performSearch(True, True)

        if mode == self.MODE_REPLACE:
            self.replaceCombo.setFocus()
        else:
            self.findtextCombo.setFocus()

        self.__subscribeToEditorSignals()

    def hide(self):
        """Overriden hide method"""
        self.__unsubscribeFromEditorSignals()
        QWidget.hide(self)

    def onQuickHighlight(self, text, wordFlag, count):
        """Called when the editor receives Ctrl+' or < or >"""
        if self.findtextCombo.isVisible():
            if text:
                self.__disconnectOnChanges()
                self.regexpCheckBox.setChecked(False)
                self.caseCheckBox.setChecked(False)
                self.wordCheckBox.setChecked(wordFlag == 1)
                self.findtextCombo.setEditText(text)
                self.__connectOnChanges()
                self.__setBackgroundColor(self.BG_NOMATCH if count == 0 else
                                          self.BG_MATCH)
                self.findPrevButton.setEnabled(count > 0)
                self.findNextButton.setEnabled(count > 0)
                self.__cursorPositionChanged()

    def __onCriteriaChanged(self, _):
        """Triggered when the search text or a checkbox state changed"""
        # All the opened buffers match cache needs to be reset to trigger
        # re-search next time the user switches the buffer
        self.editorsManager.resetTextSearchMatchCache()

        self.__performSearch(True, True)

    def __appendReplaceMessage(self):
        """Appends a proper message to the status bar after replace"""
        mainWindow = GlobalData().mainWindow
        currentMessage = mainWindow.getCurrentStatusBarMessage()
        currentIndex, totalMatches = self.__editor.getMatchesInfo()
        if totalMatches is None or totalMatches == 0:
            msg = currentMessage + '; no more matches'
        else:
            if currentIndex is not None:
                msg = currentMessage + '; match %d of %d' % (currentIndex,
                                                             totalMatches)
            else:
                if totalMatches == 1:
                    msg = currentMessage + '; 1 match left'
                else:
                    msg = currentMessage + '; %d matches left' % totalMatches
            if totalMatches > Settings()['maxHighlightedMatches']:
                msg += ' (too many to highlight)'

        mainWindow.showStatusBarMessage(msg)

    def __onReplaceAll(self):
        """Triggered when replace all button is clicked"""
        if self.replaceCombo.isVisible():
            if self.replaceAllButton.isEnabled():
                replaceText = self.replaceCombo.currentText()
                self.__editor.replaceAllMatches(replaceText)
                self.__appendReplaceMessage()
                self.__cursorPositionChanged()
                self.__updateHistory()

                count = self.__editor.getCurrentMatchesCount()
                self.__setBackgroundColor(self.BG_NOMATCH if count == 0 else
                                          self.BG_MATCH)

    # I had to use negative logic because by default the QToolButton.clicked
    # sends False as the checked status value
    def __onReplace(self, suppressMessage=False):
        """Triggered when replace current match button is clicked"""
        if self.replaceCombo.isVisible():
            if self.replaceButton.isEnabled():
                replaceText = self.replaceCombo.currentText()
                self.__editor.replaceMatch(replaceText)
                if not suppressMessage:
                    self.__appendReplaceMessage()
                self.__cursorPositionChanged()
                self.__updateHistory()

                count = self.__editor.getCurrentMatchesCount()
                self.__setBackgroundColor(self.BG_NOMATCH if count == 0 else
                                          self.BG_MATCH)

    def __onReplaceAndMove(self):
        """Triggered when replace-and-move button is clicked"""
        if self.replaceCombo.isVisible():
            if self.replaceAndMoveButton.isEnabled():
                self.__onReplace(True)
                if self.__editor.getCurrentMatchesCount() > 0:
                    self.__performSearch(False, True,
                                         self.__editor.absCursorPosition + 1,
                                         False)
                self.__appendReplaceMessage()

    def __getSearchRegexpAndFlags(self):
        """Provides the search regular expression and flags"""
        pattern = self.findtextCombo.currentText()
        pattern = pattern.replace('\u2029', '\n') # unicode paragraph -> \n

        if not self.regexpCheckBox.isChecked():
            pattern = re.escape(pattern)
        if self.wordCheckBox.isChecked():
            pattern = r'\b' + pattern + r'\b'

        flags = 0
        if not self.caseCheckBox.isChecked():
            flags = re.IGNORECASE
        return pattern, flags

    def __getRegexp(self):
        """Provides a compiled regexp"""
        pattern, flags = self.__getSearchRegexpAndFlags()
        return re.compile(pattern, flags)

    def __isSearchRegexpValid(self):
        """Compilation success and error if so"""
        pattern, flags = self.__getSearchRegexpAndFlags()
        try:
            re.compile(pattern, flags)
        except re.error as ex:
            self.__setBackgroundColor(self.BG_BROKEN)
            return False, str(ex)
        return True, None

    def __onInvalidCriteria(self, fromScratch):
        """Called when the search criteria is invalid"""
        self.__editor.resetHighlight()

        self.findNextButton.setEnabled(False)
        self.findPrevButton.setEnabled(False)
        self.replaceButton.setEnabled(False)
        self.replaceAndMoveButton.setEnabled(False)
        self.replaceAllButton.setEnabled(False)

        if not fromScratch:
            self.__editor.absCursorPosition = self.__startPoint['absPos']
            self.__editor.setFirstVisible(self.__startPoint['firstVisible'])

    def __moveToStartPoint(self):
        """Moves the editor cursor to the start point"""
        if self.__editor is not None and self.__startPoint is not None:
            self.__editor.absCursorPosition = self.__startPoint['absPos']
            self.__editor.setFirstVisible(self.__startPoint['firstVisible'])

    def __performSearch(self, fromScratch, forward,
                        absPos=None, needMessage=True):
        """Performs the incremental search"""
        if self.__editor is None:
            return

        if self.__startPoint is None:
            self.__setStartPoint()

        valid, err = self.__isSearchRegexpValid()
        if not valid:
            self.__onInvalidCriteria(fromScratch)
            GlobalData().mainWindow.showStatusBarMessage(err, 8000)
            self.__moveToStartPoint()
            return

        if self.findtextCombo.currentText() == '':
            self.__onInvalidCriteria(fromScratch)
            self.__setBackgroundColor(self.BG_IDLE)
            self.__moveToStartPoint()
            return

        # The search criteria is good, so enable the controls
        self.findNextButton.setEnabled(True)
        self.findPrevButton.setEnabled(True)

        if fromScratch:
            # Brand new editor to search in
            self.__setStartPoint()
            startPos = self.__editor.absCursorPosition
            if absPos is not None:
                startPos = absPos
            count = self.__editor.highlightRegexp(self.__getRegexp(), startPos,
                                                  forward, needMessage)
        else:
            startPos = self.__startPoint['absPos']
            if absPos is not None:
                startPos = absPos
            count = self.__editor.highlightRegexp(self.__getRegexp(), startPos,
                                                  forward, needMessage)
            if count == 0:
                self.__moveToStartPoint()
        self.__setBackgroundColor(self.BG_MATCH if count > 0
                                  else self.BG_NOMATCH)

        # The curson might not change its position so trigger the controls
        # enabling/disabling explicitly
        self.__cursorPositionChanged()

    def __onFindEnter(self, modifier):
        """Triggered when ENTER is pressed in the find combo"""
        if modifier == int(Qt.NoModifier):
            self.onNext()
        else:
            self.onPrev()

    def __onReplaceEnter(self, modifier):
        """Triggered when ENTER is pressed in the replace combo"""
        if modifier == int(Qt.NoModifier):
            self.__onReplace()
        else:
            self.__onReplaceAndMove()

    def onNext(self):
        """Triggered when the find next is clicked"""
        if self.__onPrevNext():
            self.__performSearch(False, True,
                                 self.__editor.absCursorPosition + 1)
            self.__updateHistory()

    def onPrev(self):
        """Triggered when the find prev is clicked"""
        if self.__onPrevNext():
            self.__performSearch(False, False,
                                 self.__editor.absCursorPosition - 1)
            self.__updateHistory()

    def __onPrevNext(self):
        """Checks prerequisites. Returns True if the search could be done"""
        txt = self.findtextCombo.currentText()
        if txt == "" or self.__editor is None:
            return False

        currentWidget = self.editorsManager.currentWidget()
        validWidgets = [MainWindowTabWidgetBase.PlainTextEditor,
                        MainWindowTabWidgetBase.VCSAnnotateViewer]
        return currentWidget.getType() in validWidgets

    def __setStartPoint(self):
        """Sets the new start point"""
        if self.__editor:
            self.__startPoint = {
                'absPos': self.__editor.absCursorPosition,
                'firstVisible': self.__editor.firstVisibleLine()}

    def __onfindIndexChanged(self, index):
        """Index in history has changed"""
        if index != -1:
            historyIndex = self.findtextCombo.itemData(index)
            if historyIndex is not None:
                self.__deserialize(self.__history[historyIndex])
            else:
                self.__updateHistory()

    def __subscribeToEditorSignals(self):
        """Subscribes for the cursor position notification"""
        if not self.__subscribedToCursor:
            if self.__editor:
                self.__editor.cursorPositionChanged.connect(
                    self.__cursorPositionChanged)
                self.__editor.sigHighlighted.connect(self.onQuickHighlight)
                self.__editor.textChanged.connect(self.__onTextChanged)
                self.__subscribedToCursor = True

    def __unsubscribeFromEditorSignals(self):
        """Unsubscribes from the cursor position notification"""
        if self.__subscribedToCursor:
            if self.__editor:
                try:
                    self.__editor.textChanged.disconnect(
                        self.__onTextChanged)
                    self.__editor.sigHighlighted.disconnect(
                        self.onQuickHighlight)
                    self.__editor.cursorPositionChanged.disconnect(
                        self.__cursorPositionChanged)
                except:
                    pass
            self.__subscribedToCursor = False

    def __onTextChanged(self):
        """Triggered when there are changes, i.e. the highlight is off"""
        if self.findtextCombo.isVisible():
            self.__setBackgroundColor(self.BG_IDLE)
            if self.replaceCombo.isVisible():
                self.replaceButton.setEnabled(False)
                self.replaceAndMoveButton.setEnabled(False)
                self.replaceAllButton.setEnabled(False)

    def __cursorPositionChanged(self):
        """Triggered when the cursor position is changed"""
        if self.replaceCombo.isVisible():
            onMatch = self.__editor.isCursorOnMatch()

            self.replaceButton.setEnabled(onMatch)
            self.replaceAndMoveButton.setEnabled(onMatch)

            self.replaceAllButton.setEnabled(
                self.__isCriteriaValid() and
                self.__editor.getCurrentMatchesCount() > 0)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


#
# The file was taken from eric 6 and adopted for codimension.
# Original copyright:
# Copyright (c) 2016 - 2017 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""
Breakpoints and watches on the debugger client side
"""

import os


class Breakpoint:

    """
    Breakpoint class
    """

    BREAKS = {}             # (filename, lineno) -> Breakpoint
    BREAK_IN_FILE = {}      # filename -> lineno
    BREAK_IN_FRAME_CACHE = {}

    def __init__(self, filename, lineno, temporary=False, cond=None):
        filename = os.path.abspath(filename)
        self.file = filename
        self.line = lineno
        self.temporary = temporary
        self.cond = cond
        self.enabled = True
        self.ignore = 0
        self.hits = 0
        Breakpoint.BREAKS[(filename, lineno)] = self
        lines = Breakpoint.BREAK_IN_FILE.setdefault(filename, [])
        if lineno not in lines:
            lines.append(lineno)
        Breakpoint.BREAK_IN_FRAME_CACHE.clear()

    def deleteMe(self):
        """Clears this breakpoint"""
        try:
            del Breakpoint.BREAKS[(self.file, self.line)]
            Breakpoint.BREAK_IN_FILE[self.file].remove(self.line)
            if not Breakpoint.BREAK_IN_FILE[self.file]:
                del Breakpoint.BREAK_IN_FILE[self.file]
        except KeyError:
            pass

    def enable(self):
        """Enables this breakpoint"""
        self.enabled = True

    def disable(self):
        """Disables this breakpoint"""
        self.enabled = False

    @staticmethod
    def clear_break(filename, lineno):
        """Reimplemented from bdb.py to clear a breakpoint"""
        bp = Breakpoint.BREAKS.get((filename, lineno))
        if bp:
            bp.deleteMe()
        Breakpoint.BREAK_IN_FRAME_CACHE.clear()

    @staticmethod
    def clear_all_breaks():
        """Clears all breakpoints"""
        Breakpoint.BREAKS.clear()
        Breakpoint.BREAK_IN_FILE.clear()
        Breakpoint.BREAK_IN_FRAME_CACHE.clear()

    @staticmethod
    def get_break(filename, lineno):
        """Provides the breakpoint of a particular line"""
        return Breakpoint.BREAKS.get((filename, lineno))

    @staticmethod
    def effectiveBreak(filename, lineno, frame):
        """Determine which bp for this filename:lineno is to be acted upon"""
        b = Breakpoint.BREAKS[filename, lineno]
        if not b.enabled:
            return (None, False)

        # Count every hit when bp is enabled
        b.hits += 1
        if not b.cond:
            # If unconditional, and ignoring,
            # go on to next, else break
            if b.ignore > 0:
                b.ignore -= 1
                return (None, False)

            # breakpoint and marker that's ok
            # to delete if temporary
            return (b, True)

        # Conditional bp.
        # Ignore count applies only to those bpt hits where the
        # condition evaluates to true.
        try:
            val = eval(b.cond, frame.f_globals, frame.f_locals)
            if val:
                if b.ignore > 0:
                    b.ignore -= 1
                    # continue
                else:
                    return (b, True)
            # else:
            #   continue
        except Exception:
            # if eval fails, most conservative
            # thing is to stop on breakpoint
            # regardless of ignore count.
            # Don't delete temporary,
            # as another hint to user.
            return (b, False)
        return (None, False)


class Watch:

    """
    Watch class
    """

    WATCHES = []

    def __init__(self, cond, compiledCond, flag, temporary=False):
        if cond:
            self.cond = cond
            self.compiledCond = compiledCond
            self.temporary = temporary

            self.enabled = True
            self.ignore = 0

            self.created = False
            self.changed = False
            if flag == '??created??':
                self.created = True
            elif flag == '??changed??':
                self.changed = True

            self.values = {}
            Watch.WATCHES.append(self)
        else:
            raise Exception('Inconsistency: a watch has no condition')

    def deleteMe(self):
        """Clears this watch expression"""
        try:
            del Watch.WATCHES[self]
        except ValueError:
            pass

    def enable(self):
        """Enables this watch"""
        self.enabled = True

    def disable(self):
        """Disables this watch"""
        self.enabled = False

    @staticmethod
    def clear_watch(cond):
        """Clears a watch expression"""
        try:
            Watch.WATCHES.remove(Watch.get_watch(cond))
        except ValueError:
            pass

    @staticmethod
    def clear_all_watches():
        """Clears all watch expressions"""
        del Watch.WATCHES[:]

    @staticmethod
    def get_watch(cond):
        """Provides a watch expression"""
        for b in Watch.WATCHES:
            if b.cond == cond:
                return b
        return None

    @staticmethod
    def effectiveWatch(frame):
        """Determines if a watch expression is effective"""
        for b in Watch.WATCHES:
            if not b.enabled:
                continue
            try:
                val = eval(b.compiledCond, frame.f_globals, frame.f_locals)
                if b.created:
                    if frame in b.values:
                        continue
                    b.values[frame] = [1, val, b.ignore]
                    return (b, True)
                if b.changed:
                    try:
                        if b.values[frame][1] != val:
                            b.values[frame][1] = val
                        else:
                            continue
                    except KeyError:
                        b.values[frame] = [1, val, b.ignore]

                    if b.values[frame][2] > 0:
                        b.values[frame][2] -= 1
                        continue
                    return (b, True)
                if val:
                    if b.ignore > 0:
                        b.ignore -= 1
                        continue
                    return (b, True)
            except Exception:
                continue
        return (None, False)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Detached renderer window"""


from utils.globals import GlobalData
from .qt import (QMainWindow, QTimer, QStackedWidget, QLabel, QVBoxLayout,
                 QWidget, QPalette, Qt, QFrame)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class DetachedRendererWindow(QMainWindow):

    """Detached flow ui/markdown renderer window"""

    def __init__(self, settings, em):
        QMainWindow.__init__(self, None)

        self.settings = settings
        self.em = em

        self.__widgets = QStackedWidget(self)
        self.__widgets.setContentsMargins(1, 1, 1, 1)
        self.__noRenderLabel = QLabel('\nNo rendering available for the current tab')
        self.__noRenderLabel.setFrameShape(QFrame.StyledPanel)
        self.__noRenderLabel.setAlignment(Qt.AlignHCenter)
        self.__noRenderLabel.setAutoFillBackground(True)
        font = self.__noRenderLabel.font()
        font.setPointSize(font.pointSize() + 4)
        self.__noRenderLabel.setFont(font)
        palette = self.__noRenderLabel.palette()
        palette.setColor(QPalette.Background,
                         GlobalData().skin['nolexerPaper'])
        self.__noRenderLabel.setPalette(palette)
        self.__widgets.addWidget(self.__noRenderLabel)
        self.setCentralWidget(self.__widgets)

        self.__ideClosing = False
        self.__initialisation = True

        # The size restore is done twice to avoid huge flickering
        # This one is approximate, the one in restoreWindowPosition()
        # is precise
        screenSize = GlobalData().application.desktop().screenGeometry()
        if screenSize.width() != settings['screenwidth'] or \
           screenSize.height() != settings['screenheight']:
            # The screen resolution has been changed, use the default pos
            defXPos, defYpos, \
            defWidth, defHeight = settings.getDefaultRendererWindowGeometry()
            self.resize(defWidth, defHeight)
            self.move(defXPos, defYpos)
        else:
            # No changes in the screen resolution
            self.resize(settings['rendererwidth'], settings['rendererheight'])
            self.move(settings['rendererxpos'] + settings['xdelta'],
                      settings['rendererypos'] + settings['ydelta'])


    def closeEvent(self, event):
        """Renderer is closed: explicit close via X or IDE is closed"""
        if not self.__ideClosing:
            # Update the IDE button and memorize the setting
            self.settings['floatingRenderer'] = not self.settings['floatingRenderer']
            GlobalData().mainWindow.floatingRendererButton.setChecked(False)
            self.hide()
            return

    def __registerWidget(self, widget):
        """Registers one widget basing on info from the editors manager"""
        renderLayout = QVBoxLayout()
        renderLayout.setContentsMargins(0, 0, 0, 0)
        renderLayout.setSpacing(0)
        for wid in widget.popRenderingWidgets():
            renderLayout.addWidget(wid)
        renderWidget = QWidget()
        renderWidget.setLayout(renderLayout)

        renderWidget.setObjectName(widget.getUUID())
        self.__widgets.addWidget(renderWidget)

    def show(self):
        """Overwritten show method"""
        self.__connectSignals()

        # grab the widgets
        for index in range(self.em.count()):
            widget = self.em.widget(index)
            if widget.getType() == MainWindowTabWidgetBase.PlainTextEditor:
                self.__registerWidget(widget)

        self.updateCurrent()

        QMainWindow.show(self)
        if self.__initialisation:
            self.restoreWindowPosition()

    def hide(self):
        """Overwritten hide method"""
        QMainWindow.hide(self)
        self.__disconnectSignals()

        # return widgets
        while self.__widgets.count() > 1:
            widget = self.__widgets.widget(1)
            uuid = widget.objectName()

            toBeReturned = []
            layout = widget.layout()
            for index in range(layout.count()):
                w = layout.itemAt(index).widget()
                if w is not None:
                    toBeReturned.append(w)
            for w in toBeReturned:
                layout.removeWidget(w)

            self.__widgets.removeWidget(widget)

            for index in range(self.em.count()):
                widget = self.em.widget(index)
                if widget.getUUID() == uuid:
                    widget.pushRenderingWidgets(toBeReturned)
                    break

    def __connectSignals(self):
        """Connects to all the required sugnals"""
        self.em.sigTabClosed.connect(self.__onTabClosed)
        self.em.currentChanged.connect(self.__onCurrentTabChanged)
        self.em.sigTextEditorTabAdded.connect(self.__onTextEditorTabAdded)
        self.em.sigFileTypeChanged.connect(self.__onFileTypeChanged)
        self.em.sigFileUpdated.connect(self.__onFileUpdated)
        self.em.sigBufferSavedAs.connect(self.__onBufferSavedAs)

    def __disconnectSignals(self):
        """Disconnects the signals"""
        self.em.sigBufferSavedAs.disconnect(self.__onBufferSavedAs)
        self.em.sigFileUpdated.disconnect(self.__onFileUpdated)
        self.em.sigTextEditorTabAdded.disconnect(self.__onTextEditorTabAdded)
        self.em.currentChanged.disconnect(self.__onCurrentTabChanged)
        self.em.sigTabClosed.disconnect(self.__onTabClosed)
        self.em.sigFileTypeChanged.disconnect(self.__onFileTypeChanged)

    def resizeEvent(self, resizeEv):
        """Triggered when the window is resized"""
        del resizeEv    # unused argument
        QTimer.singleShot(1, self.__resizeEventdelayed)

    def __resizeEventdelayed(self):
        """Memorizes the new window size"""
        if self.__initialisation or self.__guessMaximized():
            return

        self.settings['rendererwidth'] = self.width()
        self.settings['rendererheight'] = self.height()

    def moveEvent(self, moveEv):
        """Triggered when the window is moved"""
        del moveEv  # unused argument
        QTimer.singleShot(1, self.__moveEventDelayed)

    def __moveEventDelayed(self):
        """Memorizes the new window position"""
        if not self.__initialisation and not self.__guessMaximized():
            self.settings['rendererxpos'] = self.x()
            self.settings['rendererypos'] = self.y()

    def __guessMaximized(self):
        """True if the window is maximized"""
        # Ugly but I don't see any better way.
        # It is impossible to catch the case when the main window is maximized.
        # Especially when networked XServer is used (like xming)
        # So, make a wild guess instead and do not save the status if
        # maximized.
        availGeom = GlobalData().application.desktop().availableGeometry()
        if self.width() + abs(self.settings['xdelta']) > availGeom.width() or \
           self.height() + abs(self.settings['ydelta']) > availGeom.height():
            return True
        return False

    def restoreWindowPosition(self):
        """Makes sure that the window frame delta is proper"""
        screenSize = GlobalData().application.desktop().screenGeometry()
        if screenSize.width() != self.settings['screenwidth'] or \
           screenSize.height() != self.settings['screenheight']:
            # The screen resolution has been changed, save the new values
            self.settings['screenwidth'] = screenSize.width()
            self.settings['screenheight'] = screenSize.height()
            self.settings['xdelta'] = self.settings['xpos'] - self.x()
            self.settings['ydelta'] = self.settings['ypos'] - self.y()
            self.settings['rendererxpos'] = self.x()
            self.settings['rendererypos'] = self.y()
        else:
            # Screen resolution is the same as before
            if self.settings['rendererxpos'] != self.x() or \
               self.settings['rendererypos'] != self.y():
                # The saved delta is incorrect, update it
                self.settings['xdelta'] = self.settings['rendererxpos'] - self.x() + \
                                          self.settings['xdelta']
                self.settings['ydelta'] = self.settings['rendererypos'] - self.y() + \
                                          self.settings['ydelta']
                self.settings['rendererxpos'] = self.x()
                self.settings['rendererypos'] = self.y()
        self.__initialisation = False

    def close(self):
        """Overwritten close method. Called when the IDE is closed"""
        self.__ideClosing = True
        while self.__widgets.count() > 0:
            self.__widgets.removeWidget(self.__widgets.widget(0))
        QMainWindow.close(self)

    def __onTabClosed(self, tabUUID):
        """Triggered when the editor tab is closed"""
        for index in range(self.__widgets.count()):
            if self.__widgets.widget(index).objectName() == tabUUID:
                self.__widgets.removeWidget(self.__widgets.widget(index))
                self.updateCurrent()
                return

    def __onCurrentTabChanged(self, index):
        """Triggered when the current tab is changed"""
        del index   # :nused argument
        self.updateCurrent()

    def __onTextEditorTabAdded(self, index):
        """Triggered when a new text editor window was added"""
        widget = self.em.widget(index)
        if widget.getType() == MainWindowTabWidgetBase.PlainTextEditor:
            self.__registerWidget(widget)
        self.updateCurrent()

    def __onFileTypeChanged(self, fname, uuid, mime):
        """Triggered when a file type is changed"""
        for index in range(self.__widgets.count()):
            if self.__widgets.widget(index).objectName() == uuid:
                self.updateCurrent()
                return

    def __onBufferSavedAs(self, fname, uuid):
        """Triggered when the file was saved under another name"""
        for index in range(self.__widgets.count()):
            if self.__widgets.widget(index).objectName() == uuid:
                self.updateCurrent()
                return

    def __onFileUpdated(self, fname, uuid):
        """Triggered when the file is overwritten"""
        for index in range(self.__widgets.count()):
            if self.__widgets.widget(index).objectName() == uuid:
                self.updateCurrent()
                return

    def updateCurrent(self):
        """Updates the window title and switches to the proper widget"""
        widget = self.em.widget(self.em.currentIndex())
        if widget is None:
            # May happened when there are no widgets in the em
            return

        widgetType = widget.getType()
        if widgetType == MainWindowTabWidgetBase.PlainTextEditor:
            editor = widget.getEditor()
            isPython = editor.isPythonBuffer()
            isMarkdown = editor.isMarkdownBuffer()
            if isPython or isMarkdown:
                title = 'Floating renderer: '
                if isPython:
                    title += 'python buffer ('
                else:
                    title += 'markdown buffer ('
                title += widget.getShortName() + ')'
                self.setWindowTitle(title)

                uuid = widget.getUUID()
                for index in range(self.__widgets.count()):
                    if self.__widgets.widget(index).objectName() == uuid:
                        self.__widgets.setCurrentIndex(index)
                        break
                return

        # Not python, not markdown, i.e. no renderer
        self.__widgets.setCurrentIndex(0)
        self.setWindowTitle('Floating renderer: no renderer for the current tab')



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""codimension run parameters cache"""

# pylint: disable=W0702
# pylint: disable=W0703

import json
import logging
import os.path
from copy import deepcopy
from .runparams import RunParameters, runParamsToJSON, runParamsFromJSON
from .config import DEFAULT_ENCODING


class RunParametersCache:

    """Provides the run parameters cache"""

    def __init__(self):
        # path -> RunParameters, see runparams.py
        # The path can be relative or absolute:
        # relative for project files, absolute for non-project ones
        self.__cache = {}
        self.__rpFileName = None

    def reset(self):
        """Resets the binding to the file system"""
        self.__cache = {}
        self.__rpFileName = None

    def setup(self, dirName):
        """Binds the cache to a disk file"""
        # Just in case - flush the previous data if they were bound
        RunParametersCache.save(self)

        dirName = os.path.realpath(dirName)
        if not dirName.endswith(os.path.sep):
            dirName += os.path.sep
        if not os.path.isdir(dirName):
            raise Exception('Directory name is expected for the '
                            'run parameters cache. The given ' +
                            dirName + ' is not.')

        self.__rpFileName = dirName + 'runparams.json'
        if os.path.exists(self.__rpFileName):
            RunParametersCache.load(self)

    def load(self):
        """Loads the cache from the given file"""
        if self.__rpFileName:
            try:
                with open(self.__rpFileName, 'r',
                          encoding=DEFAULT_ENCODING) as diskfile:
                    self.__cache = json.load(diskfile,
                                             object_hook=runParamsFromJSON)
            except Exception as exc:
                logging.error('Error loading run paramaters cache '
                              '(from %s): %s', self.__rpFileName, str(exc))
                self.__cache = {}

    def save(self):
        """Saves the cache into the given file"""
        if self.__rpFileName:
            try:
                with open(self.__rpFileName, 'w',
                          encoding=DEFAULT_ENCODING) as diskfile:
                    json.dump(self.__cache, diskfile, default=runParamsToJSON)
            except Exception as exc:
                logging.error('Error saving run paramaters cache (to %s): %s',
                              self.__rpFileName, str(exc))

    def getRunParameters(self, path):
        """Provides the required parameters object"""
        try:
            return deepcopy(self.__cache[path])
        except KeyError:
            return RunParameters()

    def addRunParameters(self, path, params):
        """Adds run params into cache if needed"""
        if params.isDefault():
            self.removeRunParameters(path)
            return
        # Non-default, so need to insert
        self.__cache[path] = deepcopy(params)
        RunParametersCache.save(self)

    def removeRunParameters(self, path):
        """Removes one item from the map"""
        try:
            del self.__cache[path]
            RunParametersCache.save(self)
        except KeyError:
            return



from socket import socket, error, AF_INET, SOCK_STREAM, gethostbyname, gaierror
import sys

def createSocket():
    """Creates a socket"""
    try:
        s = socket(AF_INET, SOCK_STREAM)
    except error as err:
        print("Socket creation failed")
        print(str(err))
        return None
    except:
        print("Unknown error")
    return s

# First create a socket
s = createSocket()
if s is None:
    sys.exit(1)

try:
    hostIP = gethostbyname('www.google.com')
except gaierror:
    # Could not resolve the host
    print("Host resolving error")
    sys.exit()

# Connecting to the google server
s.connect((hostIP, 80))




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""import utility functions"""

# pylint: disable=W0702
# pylint: disable=W0703

import sys
import os
import os.path
import importlib
from cdmpyparser import getBriefModuleInfoFromMemory
from ui.qt import QApplication
from .globals import GlobalData
from .fileutils import isPythonFile


def getImportsList(fileContent):
    """Parses a python file and provides a list imports in it"""
    info = getBriefModuleInfoFromMemory(fileContent)
    return info.imports


def getImportsInLine(fileContent, lineNumber):
    """Provides a list of imports in in the given import line"""
    imports = []
    importsWhat = []
    info = getBriefModuleInfoFromMemory(str(fileContent))
    for importObj in info.imports:
        if importObj.line == lineNumber:
            if importObj.name not in imports:
                imports.append(importObj.name)
            for whatObj in importObj.what:
                if whatObj.name not in importsWhat:
                    importsWhat.append(whatObj.name)
    return imports, importsWhat


def __scanDir(prefix, path, infoLabel=None):
    """Recursive scan for modules"""
    if infoLabel is not None:
        infoLabel.setText("Scanning " + path + "...")
        QApplication.processEvents()

    result = []
    for item in os.listdir(path):
        if item in [".svn", ".cvs", ".git", ".hg"]:
            continue
        if os.path.isdir(path + item):
            result += __scanDir(prefix + item + ".",
                                path + item + os.path.sep,
                                infoLabel)
            continue

        if not isPythonFile(path + item):
            continue
        if item.startswith('__init__.'):
            if prefix != "":
                result.append(prefix[: -1])
            continue

        nameParts = item.split('.')
        result.append(prefix + nameParts[0])
    return result


def buildDirModules(path, infoLabel=None):
    """Builds a list of modules how they may appear in the import statements"""
    abspath = os.path.abspath(path)
    if not os.path.exists(abspath):
        raise Exception("Cannot build list of modules for not "
                        "existed dir (" + path + ")")
    if not os.path.isdir(abspath):
        raise Exception("Cannot build list of modules. The path " + path +
                        " is not a directory.")
    if not abspath.endswith(os.path.sep):
        abspath += os.path.sep
    return __scanDir("", abspath, infoLabel)


def __appendResult(value, result):
    """Appends to the results if it is unique"""
    if not value in result:
        result.append(value)


def __resolveImport(importObj, baseAndProjectPaths, result, errors):
    """Resolves imports like: 'import x'"""

    # import x.y
    # Could be (priority wise)
    # I:   <dir>/x/y/__init__.py
    # II:  <dir>/x/y.py

    oldSysPath = sys.path
    sys.path = GlobalData().originalSysPath + baseAndProjectPaths

    try:
        spec = importlib.util.find_spec(importObj.name)
        if spec:
            if spec.has_location:
                __appendResult([importObj.name, spec.origin, []], result)
                return
            if spec.loader is not None:
                if 'builtin' in spec.loader.__name__.lower():
                    __appendResult([importObj.name, 'built-in', []], result)
                    return

            # Something unknown; it's not clear what to do
    except:
        pass
    finally:
        sys.path = oldSysPath

    errors.append("Could not resolve 'import " + importObj.name +
                  "' at line " + str(importObj.line))


def __resolveFrom(importObj, importName, basePath, result, errors):
    """Common resolution imports like 'from [.]x import y"""
    try:
        spec = importlib.util.find_spec(importName, basePath)
        if spec:
            if spec.has_location:
                __appendResult([importObj.name, spec.origin,
                                [what.name for what in importObj.what]],
                               result)
                return

            # No location but it could be a builtin module
            if spec.loader is not None:
                if 'builtin' in spec.loader.__name__.lower():
                    __appendResult([importObj.name, 'built-in',
                                    [what.name for what in importObj.what]],
                                   result)
                    return

                # Unknown loader so not clear what to do
                errors.append("Could not resolve 'from " +
                              importObj.name + " import ..." +
                              "' at line " + str(importObj.line))
                return

            # Loader is None but found something. Maybe it is a submodule
            if spec.submodule_search_locations:
                for what in importObj.what:
                    impName = importName + '.' + what.name
                    found = False
                    try:
                        spec = importlib.util.find_spec(impName, basePath)
                        if spec:
                            visibleName = importObj.name + '.' + what.name
                            if spec.has_location:
                                __appendResult([visibleName, spec.origin, []],
                                               result)
                                found = True
                            elif spec.loader is not None:
                                if 'builtin' in spec.loader.__name__.lower():
                                    __appendResult(
                                        [visibleName, 'builtin module', []],
                                        result)
                                    found = True
                    except:
                        pass
                    if not found:
                        errors.append("Could not resolve 'from " +
                                      importObj.name + " import " +
                                      what.name + "' at line " +
                                      str(importObj.line))
                return
    except:
        pass

    errors.append("Could not resolve 'from " +
                  importObj.name + " import ...' at line " +
                  str(importObj.line))


def __resolveFromImport(importObj, basePath, baseAndProjectPaths,
                        result, errors):
    """Resolves imports like: 'from x import y'"""

    # from x.y import z
    # Could be (priority wise)
    # I:    <dir>/x/y/__init__.py  -> z
    # II:   <dir>/x/y.py  -> z
    # III:  <dir>/x/y/z/__init__.py
    # IV:   <dir>/x/y/z.py

    oldSysPath = sys.path
    sys.path = GlobalData().originalSysPath + baseAndProjectPaths

    __resolveFrom(importObj, importObj.name, basePath, result, errors)

    sys.path = oldSysPath


def __resolveRelativeImport(importObj, basePath, result, errors):
    """Resolves imports like: 'from ..x import y'"""

    # from ...x.y import z
    # Could be (priority wise)
    # I:    <dir>/x/y/__init__.py  -> z
    # II:   <dir>/x/y.py  -> z
    # III:  <dir>/x/y/z/__init__.py
    # IV:   <dir>/x/y/z.py

    if basePath is None:
        errors.append("Could not resolve 'from " + importObj.name +
                      " import ...' at line " + str(importObj.line) +
                      " because the editing buffer has not been saved yet")
    else:
        path = basePath
        current = importObj.name[1:]
        error = False
        while current.startswith('.'):
            if not path:
                error = True
                break
            current = current[1:]
            path = os.path.dirname(path)
        if error:
            errors.append("Could not resolve 'from " +
                          importObj.name + " import ...' at line " +
                          str(importObj.line))
            return

        if not path:
            path = os.path.sep  # reached the root directory

        # This is a relative import so only one path needs to be searched
        oldSysPath = sys.path
        sys.path = [path]

        __resolveFrom(importObj, current, path, result, errors)

        sys.path = oldSysPath


def resolveImports(fileName, imports):
    """Resolves a list of imports.

    fileName: the file where the imports come from
    imports: a list of the Import classes coming from the cdmpyparser module

    return: ([resolved imports], [errors])
    Each resolved import is a triple [name, path, [what imported]]
        path could be .py or .so or None or 'built-in'
    errors is a list of strings
    """
    errors = []
    result = []

    origImporterCacheKeys = set(sys.path_importer_cache.keys())
    origSysModulesKeys = set(sys.modules.keys())

    if fileName:
        basePath = os.path.dirname(fileName)    # no '/' at the end
        baseAndProjectPaths = [basePath]
    else:
        basePath = None
        baseAndProjectPaths = []

    project = GlobalData().project
    if project.isLoaded():
        for importDir in project.getImportDirsAsAbsolutePaths():
            if importDir not in baseAndProjectPaths:
                baseAndProjectPaths.append(importDir)

    for importObj in imports:
        if not importObj.what:
            # case 1: import x1, y1
            __resolveImport(importObj, baseAndProjectPaths, result, errors)
        elif not importObj.name.startswith('.'):
            # case 2: from i2 import x2, y2
            __resolveFromImport(importObj, basePath, baseAndProjectPaths,
                                result, errors)
        else:
            # case 3: from .i3 import x3, y3
            #      or from . import x4, y4
            __resolveRelativeImport(importObj, basePath, result, errors)

    importlib.invalidate_caches()

    newImporterCacheKeys = set(sys.path_importer_cache.keys())
    diff = newImporterCacheKeys - origImporterCacheKeys
    for key in diff:
        del sys.path_importer_cache[key]

    newSysModulesKeys = set(sys.modules.keys())
    diff = newSysModulesKeys - origSysModulesKeys
    for key in diff:
        del sys.modules[key]

    return result, errors


def isImportModule(info, name):
    """Returns the list of really matched modules"""
    matches = []
    for item in info.imports:
        # We are interested here in those which import a module
        if item.what:
            continue

        if item.alias == "":
            if item.name == name:
                if name not in matches:
                    matches.append(name)
        else:
            if item.alias == name:
                if item.name not in matches:
                    matches.append(item.name)
    return matches


def isImportedObject(info, name):
    """Returns a list of matched modules with the real name"""
    matches = []
    for item in info.imports:
        # We are interested here in those which import an object
        if not item.what:
            continue

        for whatItem in item.what:
            if whatItem.alias == "":
                if whatItem.name == name:
                    if name not in matches:
                        matches.append([item.name, name])
            else:
                if whatItem.alias == name:
                    if whatItem.name not in matches:
                        matches.append([item.name, whatItem.name])
    return matches



#
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy sergey.satskiy@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""VCS plugin support: status cache"""

import datetime


class VCSStatus:

    """Holds the VCS file status"""

    def __init__(self):
        self.pluginID = None        # integer
        self.indicatorID = None     # integer
        self.message = None         # string or None
        self.lastUpdate = None      # time

    def __str__(self):
        return "Plugin ID: " + str(self.pluginID) + \
            " Indicator ID: " + str(self.indicatorID) + \
            " Message: " + str(self.message) + \
            " Last update: " + str(self.lastUpdate)

    def __eq__(self, other):
        if other is None:
            return False
        return self.pluginID == other.pluginID and \
            self.indicatorID == other.indicatorID and \
            self.message == other.message

    def __ne__(self, other):
        if other is None:
            return True
        return self.pluginID != other.pluginID or \
            self.indicatorID != other.indicatorID or \
            self.message != other.message


class VCSStatusCache:

    """Caches the file statuses which came from various plugins"""

    def __init__(self):
        self.cache = {}   # path -> VCSStatus

    def getStatus(self, path):
        """Provides the status if it is in the cache"""
        if path in self.cache:
            return self.cache[path]
        return None

    def updateStatus(self, path, pluginID, indicatorID, message, callback):
        """Updates the status in the cache"""
        if path in self.cache:
            item = self.cache[path]
            item.lastUpdate = datetime.datetime.now()
            if item.pluginID != pluginID or \
               item.indicatorID != indicatorID or \
               item.message != message:
                item.pluginID = pluginID
                item.indicatorID = indicatorID
                item.message = message
                if callback:
                    callback(path, item)
            return

        item = VCSStatus()
        item.pluginID = pluginID
        item.indicatorID = indicatorID
        item.message = message
        item.lastUpdate = datetime.datetime.now()
        self.cache[path] = item

        if callback:
            callback(path, item)

    def clear(self):
        """Purges the cache"""
        self.cache = {}

    def dismissPlugin(self, pluginID, callback):
        " Removes all the certain plugin entries from the cache "
        for path, status in self.cache.items():
            if status.pluginID == pluginID:
                status.lastUpdate = None
                status.pluginID = None
                status.message = None
                oldIndicator = status.indicatorID
                status.indicatorID = None
                if oldIndicator:
                    callback(path, status)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2014-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Run/profile/debug manager"""

# pylint: disable=W0702
# pylint: disable=W0703

import os.path
import logging
import time
import uuid
from subprocess import Popen
from datetime import datetime
from editor.ioconsolewidget import IOConsoleWidget
from editor.redirectedmsg import printableTimestamp
from debugger.client.protocol_cdm_dbg import (METHOD_PROC_ID_INFO,
                                              METHOD_PROLOGUE_CONTINUE,
                                              METHOD_EPILOGUE_EXIT_CODE,
                                              METHOD_EPILOGUE_EXIT,
                                              METHOD_STDOUT, METHOD_STDERR,
                                              METHOD_STDIN, KILLED,
                                              DISCONNECTED, FAILED_TO_START,
                                              STOPPED_BY_REQUEST,
                                              SYNTAX_ERROR_AT_START,
                                              UNHANDLED_EXCEPTION)
from debugger.client.cdm_dbg_utils import getParsedJSONMessage, sendJSONCommand
from ui.runparamsdlg import RunDialog
from ui.qt import (QObject, Qt, QTimer, QDialog, QApplication, QCursor,
                   QTcpServer, QHostAddress, QAbstractSocket, pyqtSignal)
from .run import getCwdCmdEnv
from .runparams import RUN, PROFILE, DEBUG
from .procfeedback import killProcess
from .globals import GlobalData
from .settings import Settings, CLEAR_AND_REUSE, NO_REUSE
from .diskvaluesrelay import getRunParameters, addRunParams


IDE_DEBUG = False
HANDSHAKE_TIMEOUT = 15
CONNECTION_TIMEOUT = 3000

STATE_PROLOGUE = 0
STATE_RUNNING = 1

# Used from outside too
def getWorkingDir(path, params):
    """Provides the working directory"""
    if params['useScriptLocation']:
        return os.path.dirname(path)
    return params['specificDir']


class RemoteProcessWrapper(QObject):

    """Wrapper to control the remote process"""

    sigFinished = pyqtSignal(str, int)
    sigClientStdout = pyqtSignal(str, str)
    sigClientStderr = pyqtSignal(str, str)
    sigClientInput = pyqtSignal(str, str, int)
    sigIncomingMessage = pyqtSignal(str, str, object)

    def __init__(self, runManager, path, serverPort, redirected, kind):
        QObject.__init__(self)
        self.procuuid = str(uuid.uuid1())
        self.path = path
        self.redirected = redirected
        self.kind = kind
        self.state = None
        self.startTime = None
        self.finishTime = None
        self.runManager = runManager

        self.__serverPort = serverPort
        self.__clientSocket = None
        self.__proc = None

    def start(self):
        """Starts the remote process"""
        params = getRunParameters(self.path)
        if self.redirected:
            cmd, environment = getCwdCmdEnv(self.kind, self.path, params,
                                            self.__serverPort, self.procuuid)
        else:
            cmd, environment = getCwdCmdEnv(self.kind, self.path, params,
                                            self.__serverPort, self.procuuid)

        self.__proc = Popen(cmd, shell=True,
                            cwd=getWorkingDir(self.path, params),
                            env=environment)
        if self.kind == DEBUG:
            time.sleep(0.1)
            procPoll = self.__proc.poll()
            if procPoll is not None:
                msg = ' '.join(['Error starting debug session.',
                                'Process has finished with return code:',
                                str(procPoll)])
                if params['customTerminal']:
                    msg += '. Incorrect custom terminal string?'
                raise Exception(msg)

    def setSocket(self, clientSocket):
        """Called when an incoming connection has come"""
        self.__clientSocket = clientSocket
        self.state = STATE_RUNNING

        self.__connectSocket()
        self.__parseClientLine()

        # Send runnee the 'start' message
        self.__sendStart()

    def hasConnected(self):
        """True if the runnee has connected to the IDE"""
        return self.__clientSocket is not None

    def cancelPendingDebugSession(self):
        """Cancels the debug session which has not started yet due to possibly
           a bad custom terminal script. I.e. the debuggee has not come back
           to the IDE to a TCP server"""
        self.runManager.cancelPendingDebugSession(self.procuuid)

    def stop(self):
        """Kills the process"""
        self.__disconnectSocket()
        self.__kill()
        self.sigFinished.emit(self.procuuid, KILLED)

    def __connectSocket(self):
        """Connects the socket slots"""
        if self.__clientSocket:
            self.__clientSocket.readyRead.connect(self.__parseClientLine)
            self.__clientSocket.disconnected.connect(self.__disconnected)

    def __disconnectSocket(self):
        """Disconnects the socket related slots"""
        if self.__clientSocket:
            try:
                self.__clientSocket.readyRead.disconnect(
                    self.__parseClientLine)
                self.__clientSocket.disconnected.disconnect(
                    self.__disconnected)
            except:
                pass

    def __closeSocket(self):
        """Closes the client socket if so"""
        if self.__clientSocket:
            try:
                self.__clientSocket.close()
            except:
                pass
            self.__clientSocket = None

    def wait(self):
        """Waits for the process"""
        self.__closeSocket()
        if self.__proc is not None:
            try:
                self.__proc.wait()
            except:
                # E.g. wait timeout
                pass

    def waitDetached(self):
        """Needs to avoid zombies"""
        try:
            if self.__proc.poll() is not None:
                self.__proc.wait()
                return True
        except:
            return True
        return False

    def __kill(self):
        """Kills the process or checks there is no process in memory"""
        if self.__proc is not None:
            try:
                self.__proc.kill()
            except:
                pass

        childPID = self.__getChildPID()
        while childPID is not None:
            try:
                # Throws an exception if cannot kill the process
                killProcess(childPID)
            except:
                pass
            nextPID = self.__getChildPID()
            if nextPID == childPID:
                break
            childPID = nextPID

        # Here: the process killed
        self.wait()
        self.__proc = None

    def __getChildPID(self):
        """Provides the child process PID if redirected"""
        if self.__serverPort is None or self.procuuid is None:
            return None

        if self.kind == RUN:
            wrapper = os.path.join('client', 'client_cdm_run.py')
        elif self.kind == PROFILE:
            wrapper = os.path.join('client', 'client_cdm_profile.py')
        else:
            wrapper = os.path.join('client', 'client_cdm_dbg.py')
        for item in os.listdir("/proc"):
            if item.isdigit():
                try:
                    f = open("/proc/" + item + "/cmdline", "r")
                    content = f.read()
                    f.close()

                    if wrapper in content:
                        if '--port' in content:
                            if str(self.__serverPort) in content:
                                if '--procuuid' in content:
                                    if self.procuuid in content:
                                        return int(item)
                except:
                    pass
        return None

    def __disconnected(self):
        """Triggered when the client closed the connection"""
        self.__kill()
        self.sigFinished.emit(self.procuuid, DISCONNECTED)

    def __sendStart(self):
        """Sends the start command to the runnee"""
        sendJSONCommand(self.__clientSocket, METHOD_PROLOGUE_CONTINUE,
                        self.procuuid, None)

    def __sendExit(self):
        """sends the exit command to the runnee"""
        self.__disconnectSocket()
        sendJSONCommand(self.__clientSocket, METHOD_EPILOGUE_EXIT,
                        self.procuuid, None)

    def sendJSONCommand(self, method, params):
        """Sends a command to the debuggee. Used by the debugger."""
        sendJSONCommand(self.__clientSocket, method, self.procuuid, params)

    def __parseClientLine(self):
        """Parses a single line from the running client"""
        while self.__clientSocket and self.__clientSocket.canReadLine():
            try:
                method, procuuid, params, jsonStr = getParsedJSONMessage(
                    self.__clientSocket)
                del procuuid    # unused

                if IDE_DEBUG:
                    print("Process wrapper received: " + str(jsonStr))

                if method == METHOD_EPILOGUE_EXIT_CODE:
                    self.__sendExit()
                    self.sigFinished.emit(self.procuuid, params['exitCode'])
                    QApplication.processEvents()
                    continue
                if method == METHOD_STDOUT:
                    self.sigClientStdout.emit(self.procuuid, params['text'])
                    QApplication.processEvents()
                    continue
                if method == METHOD_STDERR:
                    self.sigClientStderr.emit(self.procuuid, params['text'])
                    QApplication.processEvents()
                    continue
                if method == METHOD_STDIN:
                    prompt = params['prompt']
                    echo = params['echo']
                    self.sigClientInput.emit(self.procuuid, prompt, echo)
                    QApplication.processEvents()
                    continue

                # The other messages may appear only when a code is debugged
                # so they are processed outside of a generic run manager
                self.sigIncomingMessage.emit(self.procuuid, method, params)

            except Exception as exc:
                logging.error('Failure to get a message '
                              'from a remote process: %s', str(exc))

    def userInput(self, collectedString):
        """Called when the user finished input"""
        if self.__clientSocket:
            sendJSONCommand(self.__clientSocket, METHOD_STDIN,
                            self.procuuid, {'input': collectedString})


class RemoteProcess:

    """Stores attributes of a single process"""

    def __init__(self):
        self.procWrapper = None
        self.widget = None
        self.kind = None


class RunManager(QObject):

    """Manages the external running processes"""

    # script path, output file, start time, finish time, redirected
    sigProfilingResults = pyqtSignal(str, str, str, str, bool)
    sigDebugSessionPrologueStarted = pyqtSignal(object, str, object, object)
    sigIncomingMessage = pyqtSignal(str, str, object)
    sigProcessFinished = pyqtSignal(str, int)

    def __init__(self, mainWindow):
        QObject.__init__(self)
        self.__mainWindow = mainWindow
        self.__processes = []
        self.__prologueProcesses = []

        self.__tcpServer = QTcpServer()
        self.__tcpServer.newConnection.connect(self.__newConnection)
        self.__tcpServer.listen(QHostAddress.LocalHost)

        self.__newConnectionTimer = QTimer(self)
        self.__newConnectionTimer.setSingleShot(True)
        self.__newConnectionTimer.timeout.connect(self.__onNoConnectionTimer)

        self.__waitTimer = QTimer(self)
        self.__waitTimer.setSingleShot(True)
        self.__waitTimer.timeout.connect(self.__onWaitTimer)

        self.__prologueTimer = QTimer(self)
        self.__prologueTimer.setSingleShot(True)
        self.__prologueTimer.timeout.connect(self.__onPrologueTimer)

    def __onNoConnectionTimer(self):
        """Triggered when the debug client has not connected"""
        procuuid = self.__newConnectionTimer.objectName()
        self.__onProcessFinished(procuuid, FAILED_TO_START)
        logging.error('Failed to start the script: no incoming connection')

    def cancelPendingDebugSession(self, procuuid):
        """Triggered when:
           - the user requested debugging
           - the user requested cancelling but the debuggee has not established
             a connection yet. That could be due to an incorrect custom terminal
        """
        if self.__newConnectionTimer.isActive():
            self.__newConnectionTimer.stop()
        self.__onProcessFinished(procuuid, FAILED_TO_START)

    def __newConnection(self):
        """Handles new incoming connections"""
        try:
            self.__newConnectionTimer.stop()
        except:
            pass

        clientSocket = self.__tcpServer.nextPendingConnection()
        clientSocket.setSocketOption(QAbstractSocket.KeepAliveOption, 1)
        clientSocket.setSocketOption(QAbstractSocket.LowDelayOption, 1)
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        try:
            self.__waitForHandshake(clientSocket)
        except:
            QApplication.restoreOverrideCursor()
            raise
        QApplication.restoreOverrideCursor()

    def __waitForHandshake(self, clientSocket):
        """Waits for the message with the proc ID"""
        if clientSocket.waitForReadyRead(1000):
            try:
                method, procuuid, params, jsonStr = getParsedJSONMessage(
                    clientSocket)
                if IDE_DEBUG:
                    print("Run manager (wait for handshake) received: " +
                          str(jsonStr))
                if method != METHOD_PROC_ID_INFO:
                    logging.error('Unexpected message at the handshake stage. '
                                  'Expected: %s. Received: %s',
                                  METHOD_PROC_ID_INFO, str(method))
                    self.__safeSocketClose(clientSocket)
                    return None
            except (TypeError, ValueError) as exc:
                self.__mainWindow.showStatusBarMessage(
                    'Unsolicited connection to the RunManager. Ignoring...')
                self.__safeSocketClose(clientSocket)
                return None

            procIndex = self.__getProcessIndex(procuuid)
            if procIndex is not None:
                self.__onProcessStarted(procuuid)
                self.__processes[procIndex].procWrapper.setSocket(clientSocket)
            return params

    @staticmethod
    def __safeSocketClose(clientSocket):
        """No exception socket close"""
        try:
            clientSocket.close()
        except Exception as exc:
            logging.error('Run manager safe socket close: %s', str(exc))

    def __pickWidget(self, procuuid, kind):
        """Picks the widget for a process"""
        consoleReuse = Settings()['ioconsolereuse']
        if consoleReuse == NO_REUSE:
            widget = IOConsoleWidget(procuuid, kind)
            self.__mainWindow.addIOConsole(widget, kind)
            return widget

        widget = None
        consoles = self.__mainWindow.getIOConsoles()
        for console in consoles:
            if console.kind == kind:
                procIndex = self.__getProcessIndex(console.procuuid)
                if procIndex is None:
                    widget = console
                    widget.onReuse(procuuid)
                    self.__mainWindow.onReuseConsole(widget, kind)
                    if consoleReuse == CLEAR_AND_REUSE:
                        widget.clear()
                    break
        if widget is None:
            widget = IOConsoleWidget(procuuid, kind)
            self.__mainWindow.addIOConsole(widget, kind)
        return widget

    def __updateParameters(self, path, kind):
        """Displays the dialog and updates the parameters if needed"""
        params = getRunParameters(path)
        profilerParams = None
        debuggerParams = None
        if kind == PROFILE:
            profilerParams = Settings().getProfilerSettings()
        elif kind == DEBUG:
            debuggerParams = Settings().getDebuggerSettings()
        dlg = RunDialog(path, params, profilerParams, debuggerParams,
                        kind, self.__mainWindow)
        if dlg.exec_() == QDialog.Accepted:
            addRunParams(path, dlg.runParams)
            if kind == PROFILE:
                if dlg.profilerParams != profilerParams:
                    Settings().setProfilerSettings(dlg.profilerParams)
            elif kind == DEBUG:
                if dlg.debuggerParams != debuggerParams:
                    Settings().setDebuggerSettings(dlg.debuggerParams)
            return True
        return False

    def __prepareRemoteProcess(self, path, kind):
        """Prepares the data structures to start a remote proces"""
        redirected = getRunParameters(path)['redirected']
        remoteProc = RemoteProcess()
        remoteProc.kind = kind
        remoteProc.procWrapper = RemoteProcessWrapper(
            self, path, self.__tcpServer.serverPort(), redirected, kind)
        remoteProc.procWrapper.state = STATE_PROLOGUE
        if redirected or kind == DEBUG:
            self.__prologueProcesses.append((remoteProc.procWrapper.procuuid,
                                             time.time()))
            if not self.__prologueTimer.isActive():
                self.__prologueTimer.start(1000)
        if redirected:
            remoteProc.widget = self.__pickWidget(
                remoteProc.procWrapper.procuuid, kind)
            remoteProc.widget.appendIDEMessage(
                'Starting script ' + path + '...')

            remoteProc.procWrapper.sigClientStdout.connect(
                remoteProc.widget.appendStdoutMessage)
            remoteProc.procWrapper.sigClientStderr.connect(
                remoteProc.widget.appendStderrMessage)
            remoteProc.procWrapper.sigClientInput.connect(
                remoteProc.widget.input)

            remoteProc.widget.sigUserInput.connect(self.__onUserInput)

        remoteProc.procWrapper.sigFinished.connect(self.__onProcessFinished)
        remoteProc.procWrapper.sigIncomingMessage.connect(
            self.__onIncomingMessage)
        self.__processes.append(remoteProc)
        return remoteProc

    def run(self, path, needDialog):
        """Runs the given script regardless if it is redirected"""
        if needDialog:
            if not self.__updateParameters(path, RUN):
                return

        remoteProc = self.__prepareRemoteProcess(path, RUN)
        try:
            remoteProc.procWrapper.start()
            if not remoteProc.procWrapper.redirected:
                remoteProc.procWrapper.startTime = datetime.now()
                if not self.__waitTimer.isActive():
                    self.__waitTimer.start(1000)
        except Exception as exc:
            self.__onProcessFinished(remoteProc.procWrapper.procuuid,
                                     FAILED_TO_START)
            logging.error(str(exc))

    def profile(self, path, needDialog):
        """Profiles the given script regardless if it is redirected"""
        if needDialog:
            if not self.__updateParameters(path, PROFILE):
                return

        remoteProc = self.__prepareRemoteProcess(path, PROFILE)
        try:
            remoteProc.procWrapper.start()
            if not remoteProc.procWrapper.redirected:
                remoteProc.procWrapper.startTime = datetime.now()
                if not self.__waitTimer.isActive():
                    self.__waitTimer.start(1000)
        except Exception as exc:
            self.__onProcessFinished(remoteProc.procWrapper.procuuid,
                                     FAILED_TO_START)
            logging.error(str(exc))

    def debug(self, path, needDialog):
        """Debugs the given script regardless if it is redirected"""
        if needDialog:
            if not self.__updateParameters(path, DEBUG):
                return

        remoteProc = self.__prepareRemoteProcess(path, DEBUG)

        # The run parameters could be changed by another run after the
        # debugging has started so they need to be saved per session
        self.sigDebugSessionPrologueStarted.emit(
            remoteProc.procWrapper, path,
            getRunParameters(path), Settings().getDebuggerSettings())
        try:
            remoteProc.procWrapper.start()
            if not remoteProc.procWrapper.redirected:
                self.__newConnectionTimer.setObjectName(
                    remoteProc.procWrapper.procuuid)
                self.__newConnectionTimer.start(CONNECTION_TIMEOUT)

                remoteProc.procWrapper.startTime = datetime.now()
                if not self.__waitTimer.isActive():
                    self.__waitTimer.start(1000)
        except Exception as exc:
            self.__onProcessFinished(remoteProc.procWrapper.procuuid,
                                     FAILED_TO_START)
            logging.error(str(exc))

    def killAll(self):
        """Kills all the processes if needed"""
        index = len(self.__processes) - 1
        while index >= 0:
            item = self.__processes[index]
            if item.procWrapper.redirected:
                item.procWrapper.stop()
            index -= 1

        # Wait till all the processes stopped
        count = self.__getDetachedCount()
        while count > 0:
            time.sleep(0.01)
            QApplication.processEvents()
            count = self.__getDetachedCount()

    def __getDetachedCount(self):
        """Return the number of detached processes still running"""
        count = 0
        index = len(self.__processes) - 1
        while index >= 0:
            if self.__processes[index].procWrapper.redirected:
                count += 1
            index -= 1
        return count

    def kill(self, procuuid):
        """Kills a single process"""
        index = self.__getProcessIndex(procuuid)
        if index is None:
            return
        item = self.__processes[index]
        if not item.procWrapper.redirected:
            return
        item.procWrapper.stop()

    def __getProcessIndex(self, procuuid):
        """Returns a process index in the list"""
        for index, item in enumerate(self.__processes):
            if item.procWrapper.procuuid == procuuid:
                return index
        return None

    def __getPrologueProcessIndex(self, procuuid):
        """Returns the prologue process index in the list"""
        for index, item in enumerate(self.__prologueProcesses):
            if item[0] == procuuid:
                return index
        return None

    def __onProcessFinished(self, procuuid, retCode):
        """Triggered when a redirected process has finished"""
        index = self.__getProcessIndex(procuuid)
        if index is not None:
            item = self.__processes[index]
            item.procWrapper.finishTime = datetime.now()

            needProfileSignal = False
            if retCode == KILLED:
                msg = "Script killed"
                tooltip = "killed"
            elif retCode == DISCONNECTED:
                msg = "Connection lost to the script process"
                tooltip = "connection lost"
            elif retCode == FAILED_TO_START:
                msg = "Script failed to start"
                tooltip = "failed to start"
            elif retCode == STOPPED_BY_REQUEST:
                # Debugging only: user clicked 'stop'
                msg = "Script finished by the user request"
                tooltip = "stopped by user"
                item.procWrapper.wait()
            elif retCode == UNHANDLED_EXCEPTION:
                # Debugging only: unhandled exception
                msg = "Script finished due to an unhandled exception"
                tooltip = "unhandled exception"
                item.procWrapper.wait()
            elif retCode == SYNTAX_ERROR_AT_START:
                msg = "Failure to run due to a syntax error"
                tooltip = "syntax error"
                item.procWrapper.wait()
            else:
                msg = "Script finished with exit code " + str(retCode)
                tooltip = "finished, exit code " + str(retCode)
                item.procWrapper.wait()
                if item.kind == PROFILE:
                    needProfileSignal = True

            if item.widget:
                item.widget.scriptFinished()
                item.widget.appendIDEMessage(msg)
                self.__mainWindow.updateIOConsoleTooltip(procuuid, tooltip)
                self.__mainWindow.onConsoleFinished(item.widget)
                item.widget.sigUserInput.disconnect(self.__onUserInput)

            if needProfileSignal:
                self.__sendProfileResultsSignal(item.procWrapper)

            del self.__processes[index]

        prologueProcessIndex = self.__getPrologueProcessIndex(procuuid)
        if prologueProcessIndex is not None:
            del self.__prologueProcesses[prologueProcessIndex]

        self.sigProcessFinished.emit(procuuid, retCode)

    def __onProcessStarted(self, procuuid):
        """Triggered when a process has started"""
        index = self.__getProcessIndex(procuuid)
        if index is not None:
            item = self.__processes[index]
            if item.widget:
                msg = item.widget.appendIDEMessage('Script started')
                item.procWrapper.startTime = msg.timestamp

    def __onUserInput(self, procuuid, userInput):
        """Triggered when the user input is collected"""
        index = self.__getProcessIndex(procuuid)
        if index is not None:
            item = self.__processes[index]
            if item.procWrapper.redirected:
                item.procWrapper.userInput(userInput)

    def __onWaitTimer(self):
        """Triggered when the timer fired"""
        needNewTimer = False
        index = len(self.__processes) - 1
        while index >= 0:
            item = self.__processes[index]
            if not item.procWrapper.redirected:
                if item.procWrapper.waitDetached():
                    item.procWrapper.finishTime = datetime.now()
                    if item.procWrapper.kind == PROFILE:
                        self.__sendProfileResultsSignal(item.procWrapper)
                    del self.__processes[index]
                else:
                    needNewTimer = True
            index -= 1
        if needNewTimer:
            self.__waitTimer.start(1000)

    def __sendProfileResultsSignal(self, procWrapper):
        """Sends a signal to tell that the results are available"""
        self.sigProfilingResults.emit(
            procWrapper.path,
            GlobalData().getProfileOutputPath(procWrapper.procuuid),
            printableTimestamp(procWrapper.startTime),
            printableTimestamp(procWrapper.finishTime),
            procWrapper.redirected)

    def __onIncomingMessage(self, procuuid, method, params):
        """Handles a debugger incoming message"""
        if IDE_DEBUG:
            print('Debugger message from ' + procuuid +
                  ' Method: ' + method + ' Prameters: ' + repr(params))

        self.sigIncomingMessage.emit(procuuid, method, params)

    def __onPrologueTimer(self):
        """Triggered when a prologue phase controlling timer fired"""
        needNewTimer = False
        index = len(self.__prologueProcesses) - 1
        while index >= 0:
            procuuid, startTime = self.__prologueProcesses[index]
            procIndex = self.__getProcessIndex(procuuid)
            if procIndex is None:
                # No such process anymore
                del self.__prologueProcesses[index]
            else:
                item = self.__processes[procIndex]
                if item.procWrapper.state != STATE_PROLOGUE:
                    # The state has been changed
                    del self.__prologueProcesses[index]
                else:
                    if time.time() - startTime > HANDSHAKE_TIMEOUT:
                        # Waited too long
                        item.widget.appendIDEMessage(
                            'Timeout: the process did not start; '
                            'killing the process.')
                        item.procWrapper.stop()
                    else:
                        needNewTimer = True
            index -= 1
        if needNewTimer:
            self.__prologueTimer.start(1000)

    def appendIDEMessage(self, procuuid, message):
        """Appends a message to the appropriate IO window"""
        index = self.__getProcessIndex(procuuid)
        if index is not None:
            item = self.__processes[index]
            if item.widget:
                item.widget.appendIDEMessage(message)
                return
        logging.error(message)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2012  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Debugger exceptions viewer"""

from ui.qt import Qt, pyqtSignal, QVBoxLayout, QWidget, QSplitter
from .clientexcptviewer import ClientExceptionsViewer
from .ignoredexcptviewer import IgnoredExceptionsViewer


class DebuggerExceptions(QWidget):

    """Implements the debugger context viewer"""

    sigClientExceptionsCleared = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.__createLayout()
        self.clientExcptViewer.sigClientExceptionsCleared.connect(
            self.__onClientExceptionsCleared)

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(1, 1, 1, 1)

        self.splitter = QSplitter(Qt.Vertical)

        self.ignoredExcptViewer = IgnoredExceptionsViewer(self.splitter)
        self.clientExcptViewer = ClientExceptionsViewer(
            self.splitter, self.ignoredExcptViewer)

        self.splitter.addWidget(self.clientExcptViewer)
        self.splitter.addWidget(self.ignoredExcptViewer)

        self.splitter.setCollapsible(0, False)
        self.splitter.setCollapsible(1, False)

        verticalLayout.addWidget(self.splitter)

    def clear(self):
        """Clears everything"""
        self.clientExcptViewer.clear()

    def addException(self, exceptionType, exceptionMessage, stackTrace):
        """Adds the exception to the view"""
        self.clientExcptViewer.addException(exceptionType, exceptionMessage,
                                            stackTrace)

    def isIgnored(self, exceptionType):
        """Returns True if this exception type should be ignored"""
        return self.ignoredExcptViewer.isIgnored(exceptionType)

    def setFocus(self):
        """Sets the focus to the client exception window"""
        self.clientExcptViewer.setFocus()

    def getTotalClientExceptionCount(self):
        """Provides the total number of the client exceptions"""
        return self.clientExcptViewer.getTotalCount()

    def __onClientExceptionsCleared(self):
        """Triggered when the user cleared exceptions"""
        self.sigClientExceptionsCleared.emit()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Variables viewer"""


from ui.qt import (Qt, QFrame, QVBoxLayout, QWidget, QSizePolicy, QMenu,
                   QSpacerItem, QGridLayout, QToolButton, QPushButton,
                   QToolBar, QSize)
from ui.combobox import CDMComboBox
from ui.labels import HeaderFitLabel
from utils.pixmapcache import getIcon
from utils.settings import Settings
from utils.globals import GlobalData
from .variablesbrowser import VariablesBrowser
from .varfilters import VARIABLE_FILTERS


class VariablesViewer(QWidget):

    """Implements the variables viewer for a debugger"""

    # First group of filters
    FilterGlobalAndLocal = 0
    FilterGlobalOnly = 1
    FilterLocalOnly = 2

    def __init__(self, debugger, parent=None):
        QWidget.__init__(self, parent)

        self.__debugger = debugger
        self.__browser = VariablesBrowser(debugger, self)
        self.__createLayout()

        self.setTabOrder(self.__browser, self.__execStatement)
        self.setTabOrder(self.__execStatement, self.__execButton)

        self.__updateFilter()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)
        verticalLayout.setSpacing(0)

        self.__headerLabel = HeaderFitLabel(self)
        self.__headerLabel.setText('Variables')
        self.__headerLabel.setSizePolicy(QSizePolicy.Expanding,
                                         QSizePolicy.Fixed)
        self.__headerLabel.setMinimumWidth(10)

        self.__filterMenu = QMenu(self)
        self.__showAllAct =  self.__filterMenu.addAction('Show all variables')
        self.__showAllAct.setData('showall')
        self.__filterMenu.addSeparator()
        self.__filters = []
        for title, settingName, _ in VARIABLE_FILTERS:
            action = self.__filterMenu.addAction(title)
            action.setCheckable(True)
            action.setData(settingName)
            self.__filters.append(action)
        self.__filterMenu.aboutToShow.connect(self.__filterMenuAboutToShow)
        self.__filterMenu.triggered.connect(self.__filterMenuTriggered)

        self.__filterButton = QToolButton(self)
        self.__filterButton.setIcon(getIcon('dbgvarflt.png'))
        self.__filterButton.setToolTip('Variable filter')
        self.__filterButton.setPopupMode(QToolButton.InstantPopup)
        self.__filterButton.setMenu(self.__filterMenu)
        self.__filterButton.setFocusPolicy(Qt.NoFocus)
        self.__filterButton.setFixedSize(self.__headerLabel.height(),
                                         self.__headerLabel.height())

        self.__execStatement = CDMComboBox(True)
        self.__execStatement.setSizePolicy(QSizePolicy.Expanding,
                                           QSizePolicy.Expanding)
        self.__execStatement.lineEdit().setToolTip(
            "Execute statement")
        self.__execStatement.setFixedHeight(26)
        self.__execStatement.editTextChanged.connect(
            self.__execStatementChanged)
        self.__execStatement.enterClicked.connect(self.__onEnterInExec)
        self.__execButton = QPushButton("Exec")
        self.__execButton.setEnabled(False)
        self.__execButton.setFixedHeight(26)
        self.__execButton.clicked.connect(self.__onExec)

        self.headerToolbar = QToolBar(self)
        self.headerToolbar.setIconSize(QSize(18, 18))
        self.headerToolbar.setContentsMargins(1, 1, 1, 1)
        self.headerToolbar.addWidget(self.__headerLabel)
        self.headerToolbar.addWidget(self.__filterButton)

        execLayout = QGridLayout()
        execLayout.setContentsMargins(1, 1, 1, 1)
        execLayout.setSpacing(1)
        execLayout.addWidget(self.__execStatement, 0, 0)
        execLayout.addWidget(self.__execButton, 0, 1)

        verticalLayout.addWidget(self.headerToolbar)
        verticalLayout.addWidget(self.__browser)
        verticalLayout.addLayout(execLayout)

    def __filterMenuAboutToShow(self):
        """Debug variable filter menu is about to show"""
        for flt in self.__filters:
            flt.setChecked(Settings()[flt.data()])

    def __filterMenuTriggered(self, act):
        """A filter has been changed"""
        name = act.data()
        if name == 'showall':
            for _, settingName, _ in VARIABLE_FILTERS:
                Settings()[settingName] = True
        else:
            Settings()[name] = not Settings()[name]
        self.__updateFilter()

    def updateVariables(self, areGlobals, frameNumber, variables):
        """Triggered when a new set of variables is received"""
        self.__browser.showVariables(areGlobals, variables, frameNumber)
        self.__updateHeaderLabel()

    def updateVariable(self, areGlobals, variables):
        """Triggered when a new variable has been received"""
        self.__browser.showVariable(areGlobals, variables)
        self.__updateHeaderLabel()

    def __updateHeaderLabel(self):
        """Updates the header text"""
        shown, total = self.__browser.getShownAndTotalCounts()
        if shown == 0 and total == 0:
            self.__headerLabel.setText("Variables")
        else:
            self.__headerLabel.setText("Variables (" + str(shown) +
                                       " of " + str(total) + ")")

    def __updateFilter(self):
        """Updates the current filter"""
        self.__browser.filterChanged()
        self.__updateHeaderLabel()

    def clear(self):
        """Clears the content"""
        self.__browser.clear()
        self.__updateHeaderLabel()

    def clearAll(self):
        """Clears everything including the history"""
        self.clear()
        self.__execStatement.lineEdit().setText("")
        self.__execStatement.clear()

    def __execStatementChanged(self, text):
        """Triggered when a exec statement is changed"""
        text = str(text).strip()
        self.__execButton.setEnabled(text != "")

    def __onEnterInExec(self):
        """Enter/return clicked in exec"""
        self.__onExec()

    def __onExec(self):
        """Triggered when the Exec button is clicked"""
        text = self.__execStatement.currentText().strip()
        if text != "":
            currentFrame = GlobalData().mainWindow.getCurrentFrameNumber()
            self.__debugger.remoteExecuteStatement(text, currentFrame)
            self.__debugger.remoteClientVariables(1, currentFrame)  # globals
            self.__debugger.remoteClientVariables(0, currentFrame)  # locals

    def switchControl(self, isInIDE):
        """Switches the UI depending where the control flow is"""
        self.__browser.setEnabled(isInIDE)
        self.__filterButton.setEnabled(isInIDE)

        self.__execStatement.setEnabled(isInIDE)
        if isInIDE:
            text = self.__execStatement.currentText().strip()
            self.__execButton.setEnabled(text != "")
        else:
            self.__execButton.setEnabled(False)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""list viewer base class for classes/func etc list viewers"""

import os.path
import logging
from utils.pixmapcache import getIcon
from utils.settings import Settings
from utils.project import CodimensionProject, getProjectFileTooltip
from utils.globals import GlobalData
from utils.fileutils import (getFileProperties, isPythonMime,
                             isCDMProjectMime, isImageViewable)
from utils.diskvaluesrelay import removeRecentFile, getRecentFiles
from .qt import (Qt, QSize, QTreeWidget, QTreeWidgetItem, QHeaderView, QMenu,
                 QToolButton, QWidget, QAction, QDialog, QVBoxLayout,
                 QSizePolicy, QToolBar, QApplication, QSplitter, QCursor)
from .projectproperties import ProjectPropertiesDialog
from .itemdelegates import NoOutlineHeightDelegate
from .labels import HeaderFitLabel
from .spacers import ToolBarExpandingSpacer


class RecentProjectViewItem(QTreeWidgetItem):

    """Single recent projects view item data structure"""

    def __init__(self, fileName):
        # full file name is expected
        projectName = os.path.basename(fileName .replace('.cdm3', ''))
        QTreeWidgetItem.__init__(self, ["", projectName + "   ", fileName])

        self.__isValid = True
        self.__isCurrent = False
        self.updateTooltip()

    def updateTooltip(self):
        """Updates the item tooltip"""
        fileName = self.getFilename()

        # Check that the file exists
        if not os.path.exists(fileName):
            self.__isValid = False
            self.setToolTip(0, 'Project file does not exist')
            self.setToolTip(1, 'Project file does not exist')
            self.__markBroken()
        else:
            # Get the project properties
            try:
                tooltip = getProjectFileTooltip(fileName)
                if Settings()['recentTooltips']:
                    self.setToolTip(1, tooltip)
                else:
                    self.setToolTip(1, "")
                self.setText(0, "")
                if fileName == GlobalData().project.fileName:
                    self.__markCurrent()
            except:
                # Cannot get project properties. Mark broken.
                self.__isValid = False
                self.setToolTip(0, 'Broken project file')
                self.setToolTip(1, 'Broken project file')
                self.__markBroken()
        self.setToolTip(2, self.getFilename())

    def __markBroken(self):
        """Mark the broken project with an icon"""
        self.setIcon(0, getIcon('brokenproject.png'))

    def __markCurrent(self):
        """Mark the current project with an icon"""
        self.setIcon(0, getIcon('currentproject.png'))
        self.__isCurrent = True

    def getFilename(self):
        """Provides the full project filename"""
        return str(self.text(2))

    def isValid(self):
        """True if the project is valid"""
        return self.__isValid

    def isCurrent(self):
        """True if the project is current"""
        return self.__isCurrent


class RecentFileViewItem(QTreeWidgetItem):

    """Single recent file view item data structure"""

    def __init__(self, fileName):
        # full file name is expected
        basename = os.path.basename(fileName)
        QTreeWidgetItem.__init__(self, ["", basename + "   ", fileName])

        self.__isValid = True
        self.updateIconAndTooltip()

    def updateIconAndTooltip(self):
        """Updates the item icon and tooltip if required"""
        if not os.path.exists(self.getFilename()):
            self.__markBroken()
        else:
            self.__markOK()

    def __markBroken(self):
        """Mark the file as broken"""
        self.__isValid = False
        self.setToolTip(0, 'File does not exist')
        self.setToolTip(1, 'File does not exist')
        self.setToolTip(2, self.getFilename())
        self.setIcon(0, getIcon('brokenproject.png'))

    def __markOK(self):
        """Mark the file as OK"""
        self.__isValid = True
        fileName = self.getFilename()
        mime, icon, _ = getFileProperties(fileName)
        if isPythonMime(mime):
            # The tooltip could be the file docstring
            info = GlobalData().briefModinfoCache.get(fileName)
            if info.docstring and Settings()['recentTooltips']:
                self.setToolTip(1, info.docstring.text)
            else:
                self.setToolTip(1, "")
            if info.isOK:
                self.setIcon(0, getIcon('filepython.png'))
            else:
                self.setIcon(0, getIcon('filepythonbroken.png'))
            self.setToolTip(0, "")
        elif isCDMProjectMime(mime):
            # Get the project properties
            try:
                self.setToolTip(0, "")
                tooltip = getProjectFileTooltip(fileName)
                if Settings()['recentTooltips']:
                    self.setToolTip(1, tooltip)
                else:
                    self.setToolTip(1, "")
                self.setText(0, "")
            except:
                # cannot get project properties. Mark broken.
                self.__isValid = False
                self.setToolTip(0, 'Broken project file')
                self.setToolTip(1, 'Broken project file')
            self.setIcon(0, icon)
        else:
            # Get the other file type icon
            self.setIcon(0, icon)

        self.setToolTip(2, self.getFilename())

    def getFilename(self):
        """Provides the full file name"""
        return str(self.text(2))

    def isValid(self):
        """True if the file is still valid"""
        return self.__isValid


class RecentProjectsViewer(QWidget):

    """Recent projects viewer implementation"""

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.__projectContextItem = None
        self.__fileContextItem = None

        self.__minH = None
        self.__maxH = None
        self.__splitterSizes = None

        self.upper = self.__createRecentFilesLayout()
        self.lower = self.__createRecentProjectsLayout()
        self.__createProjectPopupMenu()
        self.__createFilePopupMenu()

        layout = QVBoxLayout()
        layout.setContentsMargins(1, 1, 1, 1)
        self.splitter = QSplitter(Qt.Vertical)
        self.splitter.addWidget(self.upper)
        self.splitter.addWidget(self.lower)
        self.splitter.setCollapsible(0, False)
        self.splitter.setCollapsible(1, False)

        layout.addWidget(self.splitter)
        self.setLayout(layout)

        self.__populateProjects()
        self.__populateFiles()
        self.__updateProjectToolbarButtons()
        self.__updateFileToolbarButtons()

        # Debugging mode support
        self.__debugMode = False
        parent.debugModeChanged.connect(self.__onDebugMode)

    def setTooltips(self, _):
        """Switches the tooltips mode; _: switchOn"""
        for index in range(0, self.recentFilesView.topLevelItemCount()):
            self.recentFilesView.topLevelItem(index).updateIconAndTooltip()
        for index in range(0, self.projectsView.topLevelItemCount()):
            self.projectsView.topLevelItem(index).updateTooltip()

    def __createFilePopupMenu(self):
        """create the recent files popup menu"""
        self.__fileMenu = QMenu(self.recentFilesView)
        self.__openMenuItem = self.__fileMenu.addAction(
            getIcon('openitem.png'), 'Open', self.__openFile)
        self.__copyPathFileMenuItem = self.__fileMenu.addAction(
            getIcon('copymenu.png'),
            'Copy path to clipboard', self.__filePathToClipboard)
        self.__fileMenu.addSeparator()
        self.__delFileMenuItem = self.__fileMenu.addAction(
            getIcon('trash.png'), 'Delete from recent', self.__deleteFile)
        self.recentFilesView.setContextMenuPolicy(Qt.CustomContextMenu)
        self.recentFilesView.customContextMenuRequested.connect(
            self.__handleShowFileContextMenu)

        GlobalData().project.sigRecentFilesChanged.connect(self.__populateFiles)
        Settings().sigRecentFilesChanged.connect(self.__populateFiles)

    def __createProjectPopupMenu(self):
        """Creates the recent project popup menu"""
        self.__projectMenu = QMenu(self.projectsView)
        self.__prjLoadMenuItem = self.__projectMenu.addAction(
            getIcon('load.png'), 'Load', self.__loadProject)
        self.__projectMenu.addSeparator()
        self.__propsMenuItem = self.__projectMenu.addAction(
            getIcon('smalli.png'), 'Properties', self.__viewProperties)
        self.__prjCopyPathMenuItem = self.__projectMenu.addAction(
            getIcon('copymenu.png'),
            'Copy path to clipboard', self.__prjPathToClipboard)
        self.__projectMenu.addSeparator()
        self.__delPrjMenuItem = self.__projectMenu.addAction(
            getIcon('trash.png'), 'Delete from recent', self.__deleteProject)
        self.projectsView.setContextMenuPolicy(Qt.CustomContextMenu)
        self.projectsView.customContextMenuRequested.connect(
            self.__handleShowPrjContextMenu)

        Settings().sigRecentListChanged.connect(self.__populateProjects)
        GlobalData().project.sigProjectChanged.connect(self.__projectChanged)

    def __createRecentFilesLayout(self):
        """Creates the upper part - recent files"""
        recentFilesLabel = HeaderFitLabel(self)
        recentFilesLabel.setText("Recent files")
        recentFilesLabel.setSizePolicy(QSizePolicy.Expanding,
                                       QSizePolicy.Fixed)
        recentFilesLabel.setMinimumWidth(10)

        rfHeaderToolbar = QToolBar(self)
        rfHeaderToolbar.setContentsMargins(1, 1, 1, 1)
        rfHeaderToolbar.addWidget(recentFilesLabel)

        self.recentFilesView = QTreeWidget()
        self.recentFilesView.setAlternatingRowColors(True)
        self.recentFilesView.setRootIsDecorated(False)
        self.recentFilesView.setItemsExpandable(False)
        self.recentFilesView.setSortingEnabled(True)
        self.recentFilesView.setItemDelegate(NoOutlineHeightDelegate(4))
        self.recentFilesView.setUniformRowHeights(True)

        self.__filesHeaderItem = QTreeWidgetItem(["", "File",
                                                  "Absolute path"])
        self.recentFilesView.setHeaderItem(self.__filesHeaderItem)
        self.recentFilesView.header().setSortIndicator(1, Qt.AscendingOrder)

        self.recentFilesView.itemSelectionChanged.connect(
            self.__fileSelectionChanged)
        self.recentFilesView.itemActivated.connect(self.__fileActivated)

        # Toolbar part - buttons
        self.openFileButton = QAction(getIcon('openitem.png'),
                                      'Open the highlighted file', self)
        self.openFileButton.triggered.connect(self.__openFile)
        self.copyFilePathButton = QAction(getIcon('copymenu.png'),
                                          'Copy path to clipboard', self)
        self.copyFilePathButton.triggered.connect(self.__filePathToClipboard)
        self.trashFileButton = QAction(getIcon('delitem.png'),
                                       'Remove selected (not from the disk)',
                                       self)
        self.trashFileButton.triggered.connect(self.__deleteFile)

        self.upperToolbar = QToolBar()
        self.upperToolbar.setMovable(False)
        self.upperToolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.upperToolbar.setIconSize(QSize(16, 16))
        self.upperToolbar.setContentsMargins(0, 0, 0, 0)
        self.upperToolbar.addAction(self.openFileButton)
        self.upperToolbar.addAction(self.copyFilePathButton)
        self.upperToolbar.addWidget(ToolBarExpandingSpacer(self))
        self.upperToolbar.addAction(self.trashFileButton)

        recentFilesLayout = QVBoxLayout()
        recentFilesLayout.setContentsMargins(0, 0, 0, 0)
        recentFilesLayout.setSpacing(0)
        recentFilesLayout.addWidget(rfHeaderToolbar)
        recentFilesLayout.addWidget(self.upperToolbar)
        recentFilesLayout.addWidget(self.recentFilesView)

        upperContainer = QWidget()
        upperContainer.setContentsMargins(0, 0, 0, 0)
        upperContainer.setLayout(recentFilesLayout)
        return upperContainer

    def getRecentFilesToolbar(self):
        """Provides a reference to the recent files toolbar"""
        return self.upperToolbar

    def __createRecentProjectsLayout(self):
        """Creates the bottom layout"""
        recentProjectsLabel = HeaderFitLabel(self)
        recentProjectsLabel.setText('Recent projects')
        recentProjectsLabel.setSizePolicy(QSizePolicy.Expanding,
                                          QSizePolicy.Fixed)
        recentProjectsLabel.setMinimumWidth(10)

        self.__showHideButton = QToolButton(self)
        self.__showHideButton.setAutoRaise(True)
        self.__showHideButton.setIcon(getIcon('less.png'))
        self.__showHideButton.setFixedSize(recentProjectsLabel.height(),
                                           recentProjectsLabel.height())
        self.__showHideButton.setToolTip('Hide recent projects list')
        self.__showHideButton.setFocusPolicy(Qt.NoFocus)
        self.__showHideButton.clicked.connect(self.__onShowHide)

        self.rpHeaderToolbar = QToolBar(self)
        self.rpHeaderToolbar.setIconSize(QSize(16, 16))
        self.rpHeaderToolbar.setContentsMargins(1, 1, 1, 1)
        self.rpHeaderToolbar.addWidget(recentProjectsLabel)
        self.rpHeaderToolbar.addWidget(self.__showHideButton)

        # Toolbar part - buttons
        self.loadButton = QAction(getIcon('load.png'),
                                  'Load the highlighted project', self)
        self.loadButton.triggered.connect(self.__loadProject)
        self.propertiesButton = QAction(getIcon('smalli.png'),
                                        'Show the highlighted project '
                                        'properties', self)
        self.propertiesButton.triggered.connect(self.__viewProperties)
        self.copyPrjPathButton = QAction(getIcon('copymenu.png'),
                                         'Copy path to clipboard', self)
        self.copyPrjPathButton.triggered.connect(self.__prjPathToClipboard)
        self.trashButton = QAction(getIcon('delitem.png'),
                                   'Remove selected (not from the disk)', self)
        self.trashButton.triggered.connect(self.__deleteProject)

        self.lowerToolbar = QToolBar()
        self.lowerToolbar.setMovable(False)
        self.lowerToolbar.setAllowedAreas(Qt.TopToolBarArea)
        self.lowerToolbar.setIconSize(QSize(16, 16))
        self.lowerToolbar.setContentsMargins(0, 0, 0, 0)
        self.lowerToolbar.addAction(self.loadButton)
        self.lowerToolbar.addAction(self.propertiesButton)
        self.lowerToolbar.addAction(self.copyPrjPathButton)
        self.lowerToolbar.addWidget(ToolBarExpandingSpacer(self))
        self.lowerToolbar.addAction(self.trashButton)

        self.projectsView = QTreeWidget()
        self.projectsView.setAlternatingRowColors(True)
        self.projectsView.setRootIsDecorated(False)
        self.projectsView.setItemsExpandable(False)
        self.projectsView.setSortingEnabled(True)
        self.projectsView.setItemDelegate(NoOutlineHeightDelegate(4))
        self.projectsView.setUniformRowHeights(True)

        self.__projectsHeaderItem = QTreeWidgetItem(
            ["", "Project", "Absolute path"])
        self.projectsView.setHeaderItem(self.__projectsHeaderItem)

        self.projectsView.header().setSortIndicator(1, Qt.AscendingOrder)
        self.projectsView.itemActivated.connect(self.__projectActivated)
        self.projectsView.itemSelectionChanged.connect(
            self.__projectSelectionChanged)

        recentProjectsLayout = QVBoxLayout()
        recentProjectsLayout.setContentsMargins(0, 0, 0, 0)
        recentProjectsLayout.setSpacing(0)
        recentProjectsLayout.addWidget(self.rpHeaderToolbar)
        recentProjectsLayout.addWidget(self.lowerToolbar)
        recentProjectsLayout.addWidget(self.projectsView)

        lowerContainer = QWidget()
        lowerContainer.setContentsMargins(0, 0, 0, 0)
        lowerContainer.setLayout(recentProjectsLayout)
        return lowerContainer

    def getRecentProjectsToolbar(self):
        """Provides a reference to the projects toolbar"""
        return self.lowerToolbar

    def __projectSelectionChanged(self):
        """Handles the projects changed selection"""
        selected = list(self.projectsView.selectedItems())

        if selected:
            self.__projectContextItem = selected[0]
        else:
            self.__projectContextItem = None

        self.__updateProjectToolbarButtons()

    def __fileSelectionChanged(self):
        """Handles the files changed selection"""
        selected = list(self.recentFilesView.selectedItems())

        if selected:
            self.__fileContextItem = selected[0]
        else:
            self.__fileContextItem = None
        self.__updateFileToolbarButtons()

    def __updateProjectToolbarButtons(self):
        """Updates the toolbar buttons depending on the __projectContextItem"""
        if self.__projectContextItem is None:
            self.loadButton.setEnabled(False)
            self.propertiesButton.setEnabled(False)
            self.copyPrjPathButton.setEnabled(False)
            self.trashButton.setEnabled(False)
        else:
            enabled = self.__projectContextItem.isValid()
            isCurrentProject = self.__projectContextItem.isCurrent()

            self.propertiesButton.setEnabled(enabled)
            self.copyPrjPathButton.setEnabled(True)
            self.loadButton.setEnabled(enabled and
                                       not isCurrentProject and
                                       not self.__debugMode)
            self.trashButton.setEnabled(not isCurrentProject)

    def __updateFileToolbarButtons(self):
        """Updates the toolbar buttons depending on the __fileContextItem"""
        enabled = self.__fileContextItem is not None
        self.openFileButton.setEnabled(enabled)
        self.copyFilePathButton.setEnabled(enabled)
        self.trashFileButton.setEnabled(enabled)

    def __handleShowPrjContextMenu(self, coord):
        """Show the project item context menu"""
        self.__projectContextItem = self.projectsView.itemAt(coord)
        if self.__projectContextItem is None:
            return

        enabled = self.__projectContextItem.isValid()
        isCurrentProject = self.__projectContextItem.isCurrent()

        self.__propsMenuItem.setEnabled(enabled)
        self.__delPrjMenuItem.setEnabled(not isCurrentProject)
        # fName = self.__projectContextItem.getFilename()
        self.__prjLoadMenuItem.setEnabled(enabled and
                                          not isCurrentProject and
                                          not self.__debugMode)

        self.__projectMenu.popup(QCursor.pos())

    def __sortProjects(self):
        """Sort the project items"""
        self.projectsView.sortItems(
            self.projectsView.sortColumn(),
            self.projectsView.header().sortIndicatorOrder())

    def __sortFiles(self):
        """Sort the file items"""
        self.recentFilesView.sortItems(
            self.recentFilesView.sortColumn(),
            self.recentFilesView.header().sortIndicatorOrder())

    def __resizeProjectColumns(self):
        """Resize the projects list columns"""
        self.projectsView.header().setStretchLastSection(True)
        self.projectsView.header().resizeSections(
            QHeaderView.ResizeToContents)
        self.projectsView.header().resizeSection(0, 22)
        self.projectsView.header().setSectionResizeMode(
            0, QHeaderView.Fixed)

    def __resizeFileColumns(self):
        """Resize the files list columns"""
        self.recentFilesView.header().setStretchLastSection(True)
        self.recentFilesView.header().resizeSections(
            QHeaderView.ResizeToContents)
        self.recentFilesView.header().resizeSection(0, 22)
        self.recentFilesView.header().setSectionResizeMode(
            0, QHeaderView.Fixed)

    def __projectActivated(self, item, _):
        """Handles the double click (or Enter) on the item"""
        self.__projectContextItem = item
        self.__loadProject()

    def __fileActivated(self, item, _):
        """Handles the double click (or Enter) on a file item"""
        self.__fileContextItem = item
        self.__openFile()

    def __viewProperties(self):
        """Handles the 'view properties' context menu item"""
        if self.__projectContextItem is None:
            return
        if not self.__projectContextItem.isValid():
            return

        if self.__projectContextItem.isCurrent():
            # This is the current project - it can be edited
            project = GlobalData().project
            dlg = ProjectPropertiesDialog(project, self)
            if dlg.exec_() == QDialog.Accepted:
                importDirs = []
                for index in range(dlg.importDirList.count()):
                    importDirs.append(dlg.importDirList.item(index).text())

                scriptName = dlg.scriptEdit.text().strip()
                relativePath = os.path.relpath(scriptName,
                                               project.getProjectDir())
                if not relativePath.startswith('..'):
                    scriptName = relativePath

                mdDocFile = dlg.mdDocEdit.text().strip()
                relativePath = os.path.relpath(mdDocFile,
                                               project.getProjectDir())
                if not relativePath.startswith('..'):
                    mdDocFile = relativePath

                project.updateProperties(
                    {'scriptname': scriptName,
                     'mddocfile': mdDocFile,
                     'creationdate': dlg.creationDateEdit.text().strip(),
                     'author': dlg.authorEdit.text().strip(),
                     'license': dlg.licenseEdit.text().strip(),
                     'copyright': dlg.copyrightEdit.text().strip(),
                     'version': dlg.versionEdit.text().strip(),
                     'email': dlg.emailEdit.text().strip(),
                     'description': dlg.descriptionEdit.toPlainText().strip(),
                     'uuid': dlg.uuidEdit.text().strip(),
                     'importdirs': importDirs,
                     'encoding': dlg.encodingCombo.currentText().strip()})
        else:
            # This is not the current project - it can be viewed
            fName = self.__projectContextItem.getFilename()
            dlg = ProjectPropertiesDialog(fName, self)
            dlg.exec_()

    def __deleteProject(self):
        """Handles the 'delete from recent' context menu item"""
        if self.__projectContextItem is None:
            return

        # Removal from the visible list is done via a signal which comes back
        # from settings
        fName = self.__projectContextItem.getFilename()
        Settings().deleteRecentProject(fName)

    def __loadProject(self):
        """handles 'Load' context menu item"""
        if self.__projectContextItem is None:
            return
        if not self.__projectContextItem.isValid():
            return
        if self.__debugMode:
            return

        projectFileName = self.__projectContextItem.getFilename()

        if self.__projectContextItem.isCurrent():
            GlobalData().mainWindow.openFile(projectFileName, -1)
            return  # This is the current project, open for text editing

        QApplication.processEvents()
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        if os.path.exists(projectFileName):
            mainWin = GlobalData().mainWindow
            editorsManager = mainWin.editorsManagerWidget.editorsManager
            if editorsManager.closeRequest():
                prj = GlobalData().project
                prj.tabsStatus = editorsManager.getTabsStatus()
                editorsManager.closeAll()
                prj.loadProject(projectFileName)
                mainWin.activateProjectTab()
        else:
            logging.error("The project " +
                          os.path.basename(projectFileName) +
                          " disappeared from the file system.")
            self.__populateProjects()
        QApplication.restoreOverrideCursor()

    def __populateProjects(self):
        """Populates the recent projects"""
        self.projectsView.clear()
        for item in Settings()['recentProjects']:
            self.projectsView.addTopLevelItem(RecentProjectViewItem(item))

        self.__sortProjects()
        self.__resizeProjectColumns()
        self.__updateProjectToolbarButtons()

    def __populateFiles(self):
        """Populates the recent files"""
        self.recentFilesView.clear()
        for path in getRecentFiles():
            self.recentFilesView.addTopLevelItem(RecentFileViewItem(path))

        self.__sortFiles()
        self.__resizeFileColumns()
        self.__updateFileToolbarButtons()

    def __projectChanged(self, what):
        """Triggered when the current project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.__populateProjects()
            self.__populateFiles()
            return

        if what == CodimensionProject.Properties:
            # Update the corresponding tooltip
            items = self.projectsView.findItems(GlobalData().project.fileName,
                                                Qt.MatchExactly, 2)
            if len(items) != 1:
                logging.error("Unexpected number of matched projects: " +
                              str(len(items)))
                return

            items[0].updateTooltip()

    def __openFile(self):
        """Handles 'open' file menu item"""
        self.__fileContextItem.updateIconAndTooltip()
        fName = self.__fileContextItem.getFilename()

        if not self.__fileContextItem.isValid():
            logging.warning("Cannot open %s", fName)
            return

        mime, _, _ = getFileProperties(fName)
        if isImageViewable(mime):
            GlobalData().mainWindow.openPixmapFile(fName)
        else:
            GlobalData().mainWindow.openFile(fName, -1)

    def __deleteFile(self):
        """Handles 'delete from recent' file menu item"""
        self.removeRecentFile(self.__fileContextItem.getFilename())

    def __handleShowFileContextMenu(self, coord):
        """File context menu"""
        self.__fileContextItem = self.recentFilesView.itemAt(coord)
        if self.__fileContextItem is not None:
            self.__fileMenu.popup(QCursor.pos())

    def __filePathToClipboard(self):
        """Copies the file item path to the clipboard"""
        if self.__fileContextItem is not None:
            QApplication.clipboard().setText(
                self.__fileContextItem.getFilename())

    def __prjPathToClipboard(self):
        """Copies the project item path to the clipboard"""
        if self.__projectContextItem is not None:
            QApplication.clipboard().setText(
                self.__projectContextItem.getFilename())

    def onFileUpdated(self, fileName, _):
        """Triggered when the file is updated: python or project; _: uuid"""
        realPath = os.path.realpath(fileName)

        count = self.recentFilesView.topLevelItemCount()
        for index in range(0, count):
            item = self.recentFilesView.topLevelItem(index)

            itemRealPath = os.path.realpath(item.getFilename())
            if realPath == itemRealPath:
                item.updateIconAndTooltip()
                break

        for index in range(0, self.projectsView.topLevelItemCount()):
            item = self.projectsView.topLevelItem(index)

            itemRealPath = os.path.realpath(item.getFilename())
            if realPath == itemRealPath:
                item.updateTooltip()
                break

    def onFileTypeChanged(self, fileName, uuid, newMime):
        """Triggered when the file type is changed"""
        del uuid        # unused argument
        del newMime     # unused argument
        realPath = os.path.realpath(fileName)
        count = self.recentFilesView.topLevelItemCount()
        for index in range(0, count):
            item = self.recentFilesView.topLevelItem(index)

            itemRealPath = os.path.realpath(item.getFilename())
            if realPath == itemRealPath:
                item.updateIconAndTooltip()
                break

    def __onShowHide(self):
        """Triggered when show/hide button is clicked"""
        if self.projectsView.isVisible():
            self.__minH = self.lower.minimumHeight()
            self.__maxH = self.lower.maximumHeight()
            self.__splitterSizes = self.splitter.sizes()

            self.projectsView.setVisible(False)
            self.lowerToolbar.setVisible(False)
            self.__showHideButton.setIcon(getIcon('more.png'))
            self.__showHideButton.setToolTip('Show recent projects list')

            self.lower.setMinimumHeight(self.rpHeaderToolbar.height())
            self.lower.setMaximumHeight(self.rpHeaderToolbar.height())
        else:
            self.projectsView.setVisible(True)
            self.lowerToolbar.setVisible(True)
            self.__showHideButton.setIcon(getIcon('less.png'))
            self.__showHideButton.setToolTip('Hide recent projects list')

            self.lower.setMinimumHeight(self.__minH)
            self.lower.setMaximumHeight(self.__maxH)
            self.splitter.setSizes(self.__splitterSizes)

    def __onDebugMode(self, newState):
        """Triggered when debug mode has changed"""
        self.__debugMode = newState

        # Disable the load project button
        self.__updateProjectToolbarButtons()

    def removeRecentFile(self, fName):
        """Removes a single file from the recent files list"""
        removeRecentFile(fName)

        for index in range(self.recentFilesView.topLevelItemCount()):
            candidate = self.recentFilesView.topLevelItem(index)
            if candidate.getFilename() == fName:
                self.recentFilesView.takeTopLevelItem(index)
                return
